import multiprocessing as mp

import numpy as np
import pytest as pt

from scenario_gym.scenario_gym import ScenarioGym
from scenario_gym.trajectory import Trajectory
from scenario_gym.xosc_interface import import_scenario


@pt.fixture(scope="module")
def scenario(all_scenarios):
    """Load a single scenario."""
    return import_scenario(all_scenarios["a5e43fe4-646a-49ba-82ce-5f0063776566"])


@pt.fixture(scope="module")
def vanishing_scenario(scenario):
    """Create a scenario that vanishes."""
    scenario = scenario.copy()
    new_data = scenario.entities[1].trajectory.data.copy()
    subset = new_data[np.logical_and(new_data[:, 0] < 16.5, new_data[:, 0] > 2.0)]
    traj = Trajectory(subset)
    scenario.entities[1].trajectory = traj
    return scenario


def test_gym(scenario):
    """Rollout a single scenario."""
    gym = ScenarioGym(timestep=0.5, terminal_conditions=["max_length", "collision"])
    gym.set_scenario(scenario)
    gym.rollout()

    gym.reset_scenario()
    gym.step()
    assert np.allclose(gym.state.dt, 0.5)

    gym.timestep = 0.2
    gym.step()
    assert np.allclose(gym.state.t, gym.state.prev_t + 0.2)

    gym.rollout()
    v = gym.state.velocities[gym.state.scenario.entities[0]]
    assert (v[:2] == np.zeros(2)).all()


def test_reset_scenario(scenario):
    """Test loading and reseting a scenario."""
    gym = ScenarioGym()
    gym.set_scenario(scenario)
    assert [gym.state.poses[e] is not None for e in gym.state.poses]

    gym.set_scenario(scenario)
    assert (gym.state.scenario.entities[0].ref == "ego") and (
        gym.state.scenario.entities[1].ref == "vehicle_0"
    ), "Should be relabeled"


def test_reset_scenario_with_vanishing(vanishing_scenario):
    """Test loading and reseting a scenario."""
    gym = ScenarioGym()
    gym.set_scenario(vanishing_scenario)
    for e in vanishing_scenario.entities:
        if e.trajectory.min_t <= gym.state.t:
            assert e in gym.state.poses, f"{e.ref} should be in poses"


def test_rollout(vanishing_scenario):
    """Test rollout."""
    gym = ScenarioGym(timestep=0.1)
    gym.set_scenario(vanishing_scenario)
    gym.rollout()
    assert gym.state.scenario.entities[1] not in gym.state.poses


def test_persist(vanishing_scenario):
    """Test rollout."""
    gym = ScenarioGym(timestep=0.1, persist=True)
    gym.set_scenario(vanishing_scenario)
    assert len(gym.state.poses) == len(vanishing_scenario.entities), (
        "Entities not found in poses: ",
        str(
            [e.ref for e in vanishing_scenario.entities if e not in gym.state.poses]
        ),
    )
    while not gym.state.is_done:
        gym.step()
        assert len(gym.state.poses) == len(vanishing_scenario.entities), (
            "Entities not found in poses: ",
            str(
                [
                    e.ref
                    for e in vanishing_scenario.entities
                    if e not in gym.state.poses
                ]
            ),
        )


def test_render(scenario, vanishing_scenario, all_scenarios):
    """Test the rendering of the gym."""
    gym = ScenarioGym()
    gym.set_scenario(scenario)
    gym.rollout(
        render=True,
        video_path=(
            all_scenarios[scenario.name]
            .replace("Scenarios", "Recordings")
            .replace(".xosc", ".mp4")
        ),
    )
    gym.set_scenario(vanishing_scenario)
    gym.rollout(
        render=True,
        video_path=(
            all_scenarios[scenario.name]
            .replace("Scenarios", "Recordings")
            .replace(".xosc", "_vanishing.mp4")
        ),
    )


def test_manual_rollout(all_scenarios):
    """Test rollout manually."""
    scenario_path = all_scenarios["a5e43fe4-646a-49ba-82ce-5f0063776566"]

    gym = ScenarioGym(timestep=0.2)
    gym.load_scenario(scenario_path)
    gym.render()
    while not gym.state.is_done:
        gym.step()
        gym.render()
    gym.close()


def test_run_scenarios(all_scenarios):
    """Test using the run scenarios class method."""
    scenario_path = all_scenarios["a5e43fe4-646a-49ba-82ce-5f0063776566"]
    ScenarioGym.run_scenarios(
        [scenario_path, scenario_path],
        timestep=0.075,
    )


def _render_scenario(scenario, path):
    """Render one scenario."""
    gym = ScenarioGym(timestep=0.075)
    gym.set_scenario(scenario, scenario_path=path)
    gym.rollout(render=True)


def test_multi_process_scenarios(all_scenarios):
    """Test running scenarios in multiple processes."""
    scenarios = []
    num_processes = 4
    for _, path in zip(range(num_processes), all_scenarios.values()):
        scenarios.append(import_scenario(path))

    with mp.Pool(num_processes) as p:
        p.starmap(_render_scenario, zip(scenarios, all_scenarios.values()))



import os
from collections import defaultdict
from functools import lru_cache
from typing import Dict, List, Optional, Tuple, Type

from lxml import etree
from lxml.etree import Element
from scenariogeneration import xosc

from scenario_gym.catalog_entry import Catalog, CatalogEntry
from scenario_gym.entity import DEFAULT_ENTITY_TYPES, Entity


def load_object(
    entry: Element,
    catalog: Optional[Catalog] = None,
    entity_types: Tuple[Type[Entity]] = DEFAULT_ENTITY_TYPES,
    catalog_objects: Optional[List[Type[CatalogEntry]]] = None,
) -> Optional[Entity]:
    """Try to load a catalog entry with given catalog objects."""
    if catalog_objects is None:
        catalog_objects = [Ent._catalog_entry_type() for Ent in entity_types]
    for Ent, Obj in zip(entity_types, catalog_objects):
        types = Obj.xosc_names if Obj.xosc_names is not None else [Obj.__name__]
        if entry.tag in types:
            obj = Obj.from_xml(entry, catalog=catalog)
            return Ent(obj)


@lru_cache(maxsize=None)
def read_catalog(
    catalog_file: str,
    entity_types: Optional[Tuple[Type[Entity]]] = None,
) -> Tuple[Catalog, Dict[str, Entity]]:
    """
    Read a catalog and return it's name and a dictionary of entities.

    Parameters
    ----------
    catalog_file : str
        Filepath of the catalog file.

    entity_types : Optional[Tuple[Type[CatalogObject]]]
        Tuple of extra subclasses of CatalogObject that will be used when reading
        catalogs. Can be used for reading custom objects from catalog files. Must
        be immutable or lru_cache

    Returns
    -------
    catalog : Catalog
        The catalog.

    entries : Dict[str, Entity]
        A dictionary mapping each entry name to an entity with that catalog entry.
        These can then be cloned to create entities using the chosen catalog entry.

    """
    if entity_types is None:
        entity_types = DEFAULT_ENTITY_TYPES
    else:
        entity_types = entity_types + DEFAULT_ENTITY_TYPES

    catalog_objects = [Ent._catalog_entry_type() for Ent in entity_types]

    et = etree.parse(catalog_file)
    osc_root = et.getroot()
    catalog_element = osc_root.find("Catalog")
    try:
        catalog_group_name = catalog_file.split(os.sep)[-3]
    except IndexError:
        catalog_group_name = "Catalog"
    catalog = Catalog(catalog_element.attrib["name"], catalog_group_name)
    entries = {}
    for element in catalog_element.getchildren():
        entry = load_object(
            element,
            catalog=catalog,
            entity_types=entity_types,
            catalog_objects=catalog_objects,
        )
        if entry is None:
            entry = Entity(CatalogEntry.from_xml(element, catalog=catalog))
        entries[entry.catalog_entry.catalog_entry] = entry
    return catalog, entries


def write_catalogs(dirname: str, entries: List[CatalogEntry]) -> None:
    """Create catalog files for a list of catalog entries."""
    c_to_e = defaultdict(list)
    for entry in entries:
        c_to_e[entry.catalog].append(entry)

    group_to_c = defaultdict(list)
    for catalog in c_to_e:
        group_to_c[catalog.group_name].append(catalog)

    for group_name, catalogs in group_to_c.items():
        os.mkdir(os.path.join(dirname, group_name))
        for catalog in catalogs:
            c_entries = c_to_e[catalog]
            catalog_type = f"{c_entries[0].catalog_type}Catalogs"
            catalog_dir = os.path.join(dirname, group_name, catalog_type)
            if not os.path.exists(catalog_dir):
                os.mkdir(catalog_dir)

            catalog_file = os.path.join(catalog_dir, f"{catalog.name}.xosc")

            catalog_obj = xosc.CatalogFile()
            catalog_obj.create_catalog(
                catalog_file,
                catalog.name,
                (
                    f"This {c_entries[0].catalog_type}.lower() catalog was "
                    "generated by scenario_gym."
                ),
                "scenario_gym",
            )

            for entry in c_entries:
                catalog_obj.add_to_catalog(entry.to_xosc())

            catalog_obj.dump()



from dataclasses import dataclass

from scenario_gym.action import Action


@dataclass
class PedestrianAction(Action):
    """A speed and heading angle update for pedestrian agents."""

    speed: float
    heading: float



from contextlib import suppress
from functools import lru_cache
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from lxml.etree import Element
from shapely.geometry import Polygon
from shapely.strtree import STRtree

try:
    from functools import cached_property
except ImportError:

    def cached_property(fn):
        """Replace cached_property with a size 1 lru cache."""
        return property(lru_cache(maxsize=1)(fn))


try:
    from numpy.typing import ArrayLike, NDArray
except ImportError:
    ArrayLike = NDArray = np.ndarray


ArgsKwargs = Tuple[Tuple[Any, ...], Dict[str, Any]]


def detect_geom_collisions(
    geoms: List[Polygon],
    others: Optional[List[Polygon]] = None,
) -> Dict[Polygon, List[Polygon]]:
    """
    Detect collisions between polygons.

    Parameters
    ----------
    geoms : List[Polygon]
        The geometries to use.

    others : Optional[List[Polygon]]
        Additional geometries to include when checking for collisions
        with each geometry in geoms.

    Returns
    -------
    Dict[Polygon, List[Polygon]]
        A dictionary that maps each polygon in geoms to the polygons which it
        intersects.

    """
    all_geoms = geoms if others is None else geoms + others
    tree = STRtree(all_geoms)
    return {
        g: [
            g_prime
            for g_prime in tree.geometries.take(
                tree.query(g, predicate="intersects").tolist()
            )
            if g != g_prime
        ]
        for g in geoms
    }


def load_properties_from_xml(
    element: Element,
) -> Tuple[Dict[str, Union[str, float]], List[str]]:
    """
    Load properties from the xml element.

    These can be either `Property` or `File` elements. `Property` elements are
    given with `name` and `value` attributes (`name` must be unique for the
    entry) and are returned as a dict of values indexed by `name`. The value
    will be converted to a float if it can otherwise the string will be
    returned. `File` elements must have a `filepath` attribute which will
    be parsed. Multiple files can be stored with one entity.

    Returns
    -------
    properties : Dict[str, Union[str, float]]
        A dictionary of properties indexed by name.

    files : List[str]
        A list of filepaths for external files.

    """
    files = []
    properties = {}
    prop = element.find("Properties")
    if prop is not None:
        for child in prop.findall("Property"):
            try:
                v = child.attrib["value"]
                with suppress(ValueError):
                    v = float(v)
                properties[child.attrib["name"]] = v
            except KeyError as e:
                raise RuntimeError(
                    "Property could not be loaded without `value` key."
                ) from e
        for file in prop.findall("File"):
            files.append(file.attrib["filepath"])
    return properties, files



from typing import Dict, Iterable, List, Tuple

import numpy as np
from numpy.linalg import norm


def inverse_direction(vector: Iterable, normalised: bool = True) -> List[float]:
    """
    Return the inverse of a 2D vector, Iterable -> Iterable.

    Uses clockwise-rotating sign convention: (x, y) --> (y, -x)

    Optional parameter normalised (default True):
    --  True: returns vector of unit length
    --  False: returns vector of same length as the input vector
    """
    assert len(vector) == 2, "Invalid vector dimension: {0}".format(len(vector))
    if normalised:
        n = norm([vector[1], vector[0]])
        return [vector[1] / n, -vector[0] / n]
    return [vector[1], -vector[0]]


def coord_change(
    vector: List[float],
    direction: List[float],
    centre: Tuple[float, float] = (0, 0),
) -> List[float]:
    """
    Change vector coordinates to new frame of reference.

    Uses Galilean transformation to change the components of <vector> to a new
    coordinate system defined by an origin at <centre> and with components parallel
    and perpendicular to <direction>

    Optional argument centre (default [0, 0])
    """
    assert len(vector) == 2, (
        "coord_change is implemented to work in 2D. Passed vector dimension: "
        + str(len(vector))
    )
    vector = np.array(vector)
    centre = np.array(centre)
    inv_dir = inverse_direction(direction)
    return [np.dot((vector - centre), inv_dir), np.dot(vector - centre, direction)]


def acceleration(
    entity_poses: List[List[float]],
    dt: float,
    parallel_velocity: bool = False,
    i: int = 0,
) -> List[float]:
    """
    Find entity acceleration from three consecutive poses.

    Optional argument parallel_velocity (default False):
    --  True: acceleration is resolved in coords para and perp to velocity
    --  False: acceleration remains in pose coords
    """
    try:
        entity_pose_2 = entity_poses[i + 2][1:3]
        entity_pose_1 = entity_poses[i + 1][1:3]
        entity_pose_0 = entity_poses[i][1:3]
    except IndexError:
        # If too few poses received, default with accel = [0, 0]
        return [0.0, 0.0]
    velocity_1 = (entity_pose_1 - entity_pose_2) / dt
    velocity_0 = (entity_pose_0 - entity_pose_1) / dt
    accel = np.array((velocity_0 - velocity_1) / dt)
    if not parallel_velocity:
        return accel
    # Resolve acceleration vector into [para to velocity, perp to velocity]
    return [
        np.dot(velocity_1, accel) / norm(velocity_1),
        np.dot([-velocity_1[1], velocity_1[0]] / norm(velocity_1), accel),
    ]


def ahead(ego: Dict, haz: Dict) -> bool:
    """
    Determine if ego is ahead of hazard.

    Takes two entities (ego and hazard) and returns True if the ego is ahead
    of the hazard when using ego's preferential frame of reference such that:
    --  ego is at position [0.0, 0.0]
    --  ego direction is [0, 1] ...
    --  ... requiring ego velocity to be parallel to y axis
    """
    ego_position = ego["position"][1]
    haz_position = haz["position"][1]
    return ego_position > haz_position


def direction(heading: float) -> list:
    """
    Turn heading into normalised direction vector.

    angle (rad) --> [x component, y component]
    Where 0 rad corresponds to positive x unit vector, and a positively
    increasing angle moves in counter-clockwise direction.
    """
    return [np.cos(heading), np.sin(heading)]



import random
from itertools import chain
from typing import Dict, List, Optional, Tuple

import numpy as np

from scenario_gym.road_network import RoadNetwork


class RouteFinder:
    """Find routes along walkable areas in the road network."""

    def __init__(self, rn: RoadNetwork):
        """Construct the graph representation."""
        self.rn = rn
        (
            self.graph,
            self.node_to_idx,
            self.node_data,
        ) = make_pedestrian_connection_graph(rn)

    def find_route(
        self, start: np.ndarray, finish: np.ndarray
    ) -> Optional[List[Tuple[float, float]]]:
        """Find the route or return None if one cannot be found."""
        return find_route(self.graph, self.node_data, start, finish)

    def generate_route(
        self,
        n: int,
        start: Optional[np.ndarray] = None,
        no_repeat: bool = False,
    ) -> List[Tuple[float, float]]:
        """Generate a route by a random walk."""
        if start is not None:
            n0 = min(
                self.node_data,
                key=lambda x: np.linalg.norm(self.node_data[x] - start),
            )
            route = [n0]
        else:
            route = [random.choice(self.graph.keys())]
        while len(route) < n:
            suc = self.graph[route[-1]]
            if no_repeat:
                suc = list(set(suc).difference(route))
            if not suc:
                break
            route.append(random.choice(suc))
        return [self.node_data[i] for i in route]


def make_pedestrian_connection_graph(
    rn: RoadNetwork,
) -> Tuple[Dict[int, List[int]], Dict[str, int], Dict[int, Tuple[float, float]]]:
    """
    Build a graph representing the walkable surface of the road network.

    Nodes are positions in the pavements and crossings with edges connecting
    nodes if pedestrians can walk between them.

    Parameters
    ----------
    rn : RoadNetwork
        The road network.

    Returns
    -------
    graph : Dict[int, List[int]]
        The graph represented as a dictionary with integer node indexes and
        lists of neighbours of each node as values.

    node_to_idx : Dict[str, int]
        A dictionary giving a string identifier of each node index. The string
        identifier is '{road_object_id}_{index_in_object}'.

    node_data : Dict[int, Tuple[float, float]]
        A dictionary giving the xy coordinates of each node.

    """
    graph = {}
    node_to_idx = {}
    node_data = {}
    pavement_coords = {}
    for p in rn.pavements:
        pavement_coords[p.id] = np.array(
            [
                np.array(p.center.interpolate(x, normalized=False).xy).squeeze()
                for x in np.linspace(0.0, p.center.length, int(p.center.length))
            ]
        )

    crossing_coords = {}
    for c in rn.crossings:
        crossing_coords[c.id] = np.array(
            [
                np.array(c.center.interpolate(x, normalized=False).xy).squeeze()
                for x in np.linspace(0.0, c.center.length, int(c.center.length))
            ]
        )

    for obj, coords in chain(pavement_coords.items(), crossing_coords.items()):
        for i, (x, y) in enumerate(coords):
            node_to_idx[f"{obj}_{i}"] = len(node_to_idx)
            graph[node_to_idx[f"{obj}_{i}"]] = []
            node_data[node_to_idx[f"{obj}_{i}"]] = (x, y)

    for obj, coords in chain(pavement_coords.items(), crossing_coords.items()):
        for i in range(len(coords) - 1):
            graph[node_to_idx[f"{obj}_{i}"]].append(node_to_idx[f"{obj}_{i+1}"])
            graph[node_to_idx[f"{obj}_{i+1}"]].append(node_to_idx[f"{obj}_{i}"])

    for c in rn.crossings:
        for p in c.pavements:
            c_coords, p_coords = crossing_coords[c.id], pavement_coords[p]
            c_idx, p_idx = np.unravel_index(
                np.linalg.norm(
                    c_coords[:, None, :] - p_coords[None, :, :], axis=-1
                ).argmin(),
                (c_coords.shape[0], p_coords.shape[0]),
            )
            graph[node_to_idx[f"{c.id}_{c_idx}"]].append(
                node_to_idx[f"{p}_{p_idx}"]
            )
            graph[node_to_idx[f"{p}_{p_idx}"]].append(
                node_to_idx[f"{c.id}_{c_idx}"]
            )
    return graph, node_to_idx, node_data


def shortest_path(
    graph: Dict[int, List[int]],
    start: int,
    goal: int,
) -> Optional[List[int]]:
    """
    Find the shortest path between two nodes in the graph.

    Returns
    -------
    path : Optional[List[int]]
        The path as a list of nodes. Will return None if the start and goal are
        not path connected.

    """
    explored = []
    queue = [[start]]
    if start == goal:
        return [start]
    while queue:
        path = queue.pop(0)
        node = path[-1]
        if node not in explored:
            neighbours = graph[node]
            for neighbour in neighbours:
                new_path = path.copy()
                new_path.append(neighbour)
                queue.append(new_path)
                if neighbour == goal:
                    return new_path
            explored.append(node)


def find_route(
    graph: Dict[int, List[int]],
    node_data: Dict[int, Tuple[float, float]],
    start: np.ndarray,
    finish: np.ndarray,
) -> Optional[np.ndarray]:
    """
    Find the shortest path in the walkable area given by the graph.

    First finds the closest nodes to the start and finish then connects
    them with the shortest path in the graph. Will return None if the
    start and finish are not path connected.
    """
    if not node_data:
        return np.array([start] + [finish])
    start_node = min(
        node_data, key=lambda n: np.linalg.norm(np.array(node_data[n]) - start)
    )
    end_node = min(
        node_data,
        key=lambda n: np.linalg.norm(np.array(node_data[n]) - finish),
    )
    route = shortest_path(graph, start_node, end_node)
    if route is None:
        return None
    xy = [list(node_data[n]) for n in route]
    return np.array([start] + xy + [finish])



from dataclasses import dataclass, field
from random import choice
from typing import Dict, Optional, Tuple

import numpy as np
from nuscenes import NuScenes
from nuscenes.prediction import PredictHelper
from nuscenes.prediction.input_representation.static_layers import load_all_maps
from scipy.spatial.transform import Rotation
from shapely.geometry import LineString

from scenario_gym.catalog_entry import BoundingBox, Catalog, CatalogEntry
from scenario_gym.entity import Entity
from scenario_gym.road_network import Lane, LaneType, Road, RoadNetwork
from scenario_gym.scenario import Scenario
from scenario_gym.trajectory import Trajectory


@dataclass
class NuScenesInstanceData:
    """Class for storing data for NuScenes "instances" (entities)."""

    category_name: str
    trajectory: list = field(default_factory=list)
    times: list = field(default_factory=list)
    sizes: list = field(default_factory=list)
    rotations: list = field(default_factory=list)


class Catalogs:
    """
    Catalogs used in the nuScenes dataset.

    Using the dimensions defined "at runtime" from the bounding box data.
    """

    nuScenes_catalog = Catalog("nuScenesCatalog", None)


class NuScenesImporter:
    """
    Class for importing nuScenes scenes into scenario gym scenarios.

    Parameters
    ----------
    data_root : str
        Dataset root path to be passed into the NuScenes constructor.
    dataset : str, optional
        Dataset name to be passed into the NuScenes constructor, by default
        "v1.0-mini"
    map_radius_multiplier : float, optional
        When including a map around a scene, a centrepoint will be computed as the
        average position of all trajectories; a radius will be computed as the max
        x and y range of coordinates. All lanes within this radius multiplied by
        the map_radius_multiplier of the centrepoint will be included. By default
        1.5
    pre_loaded_data : NuScenes, optional
        Pre-loaded (indexed) data object, for much faster init if the NuScenes
        dataset has been pre-loaded already, by default None.

    """

    def __init__(
        self,
        data_root: str,
        dataset: str = "v1.0-mini",
        map_radius_multiplier: float = 1.5,
        pre_loaded_data: Optional[NuScenes] = None,
    ):
        self.data_root = data_root
        self.dataset = dataset

        if pre_loaded_data is not None:
            self.data = pre_loaded_data
        else:
            self.data = NuScenes(self.dataset, dataroot=self.data_root)
        self.predict_helper = PredictHelper(self.data)

        self.maps = load_all_maps(self.predict_helper)

        self.map_radius_multiplier = map_radius_multiplier

    def _convert_nuScenes_map_to_road_network(
        self, map_name, centre_coordinate: np.ndarray, map_radius: float
    ) -> RoadNetwork:
        map = self.maps[map_name]

        lane_records = map.get_records_in_radius(
            *centre_coordinate, map_radius, ["lane", "lane_connector"]
        )

        roads = []
        lane_centres = map.discretize_lanes(
            lane_records["lane"], 0.1
        ) | map.discretize_lanes(lane_records["lane_connector"], 0.1)

        lane_keys_and_records = [("lane", l) for l in lane_records["lane"]] + [
            ("lane_connector", l) for l in lane_records["lane_connector"]
        ]
        lane_ids = set([l[1] for l in lane_keys_and_records])
        for lane_key, lane_record in lane_keys_and_records:
            lane = map.get(lane_key, lane_record)
            bounding_poly = map.extract_polygon(lane["polygon_token"])
            lane_centre = LineString(np.array(lane_centres[lane_record])[:, :2])

            sg_lane = Lane(
                lane_record,
                bounding_poly,
                lane_centre,
                [
                    l_id
                    for l_id in map.get_outgoing_lane_ids(lane_record)
                    if l_id in lane_ids
                ],
                [
                    l_id
                    for l_id in map.get_incoming_lane_ids(lane_record)
                    if l_id in lane_ids
                ],
                LaneType.driving,
                elevation=np.array(lane_centres[lane_record]),
            )

            roads.append(
                Road(
                    f"road_{lane_key}_{lane_record}",
                    bounding_poly,
                    lane_centre,
                    [sg_lane],
                )
            )

        road_network = RoadNetwork(
            roads=roads,
            intersections=[],
        )

        return road_network

    def convert_instance_sample_token_to_gym(
        self,
        ego_instance_token: str,
        sample_token: str,
        seconds_history: float = 2.0,
        seconds_future: float = 6.0,
    ) -> Scenario:
        """
        Convert an (instance token, sample token) pair to a scenario gym scenario.

        Note in the resulting scenario, the sample token will occur at t==0. Since
        rendering begins at t==0, to render the entire scenario, translate all
        entities in time.

        Parameters
        ----------
        ego_instance_token : str
            Instance token. This instance (entity) will be treated as the ego.
        sample_token : str
            Sample token to treat as t == 0. Later and earlier sample tokens will
            be queried to create the scenario.
        seconds_history : float, optional
            Seconds before the provided sample_token to query samples for, by
            default 2.0
        seconds_future : float, optional
            Seconds after the provided sample_token to query samples for, by
            default 6.0

        Returns
        -------
        Scenario
            Scenario gym scenario corresponding to this NuScenes data.

        """
        # Link from instance IDs to relevant instance (entity) data
        # We will use the current sample token as t = 0.
        # Note scenario simulation starts at t=0, so to simulate the whole scenario
        # including past, the scenario.translate method should be used.
        instance_token_to_data: Dict[str, NuScenesInstanceData] = {}

        instance_token_to_past_data = self.predict_helper.get_past_for_sample(
            sample_token,
            seconds=seconds_history,
            in_agent_frame=False,
            just_xy=False,
        )
        instance_token_to_current_data = {
            d["instance_token"]: [d]
            for d in self.predict_helper.get_annotations_for_sample(sample_token)
        }
        instance_token_to_future_data = self.predict_helper.get_future_for_sample(
            sample_token,
            seconds=seconds_future,
            in_agent_frame=False,
            just_xy=False,
        )

        for instance_token in (
            instance_token_to_past_data.keys()
            | instance_token_to_future_data.keys()
            | instance_token_to_current_data.keys()
        ):
            past_data = instance_token_to_past_data.get(instance_token, [])
            current_data = instance_token_to_current_data.get(instance_token, [])
            future_data = instance_token_to_future_data.get(instance_token, [])

            past_times = np.linspace(
                -0.5,
                -0.5 * (len(past_data)),
                len(past_data),
            )
            future_times = np.linspace(
                0.5,
                0.5 * (len(future_data)),
                len(future_data),
            )

            combined_times = list(past_times) + [0.0] + list(future_times)
            combined_data = past_data + current_data + future_data

            assert len(combined_data) == len(combined_times)

            trajectory = [annotation["translation"] for annotation in combined_data]
            sizes = [annotation["size"] for annotation in combined_data]
            rotations = [annotation["rotation"] for annotation in combined_data]

            instance_token_to_data[instance_token] = NuScenesInstanceData(
                combined_data[0]["category_name"],
                trajectory=trajectory,
                times=combined_times,
                sizes=sizes,
                rotations=rotations,
            )

        map_name = self.predict_helper.get_map_name_from_sample_token(sample_token)

        entities, road_network = self._convert_to_entities_road_network(
            instance_token_to_data, map_name, ego_instance_token=ego_instance_token
        )

        scenario = Scenario(
            entities,
            name="_".join(((ego_instance_token, sample_token))),
            road_network=road_network,
        )
        return scenario

    def convert_scene_to_gym(
        self, scene_token: str, ego_instance_token: Optional[str] = None
    ) -> Scenario:
        """
        Convert a complete nuScenes scene to a scenario gym scenario.

        Where ego_instance_token is provided, the instance (agent) which corresponds
        to that token will be treated as the ego. Otherwise, a random car will be
        chosen.

        See https://www.nuscenes.org/nuscenes#data-format for nuScenes schema.

        Parameters
        ----------
        scene_token : str
            Unique identifier for a nuScenes scene
        ego_instance_token : Optional[str], optional
            Identifier for the instance to be used as the scenario gym ego, by
            default None

        Returns
        -------
        Scenario
            Converted scenario gym scenario corresponding to this nuScenes scene.

        Raises
        ------
        KeyError
            If the provided ego_instance_token is not found in the scene.
        ValueError
            If no ego_instance_token is provided and there are no cars in the scene
            to be chosen as the ego.

        """
        scene_data = self.data.get("scene", scene_token)

        sample_annotations: list[list[dict]] = []

        first_sample_token = scene_data["first_sample_token"]
        last_sample_token = scene_data["last_sample_token"]
        current_sample_token = first_sample_token

        while current_sample_token != last_sample_token:
            if current_sample_token is None or current_sample_token == "":
                print(
                    "WARNING: Got an unexpected sample token of "
                    + str(current_sample_token)
                )
                break
            sample_annotations.append(
                self.predict_helper.get_annotations_for_sample(current_sample_token)
            )
            current_sample_token = self.data.get("sample", current_sample_token)[
                "next"
            ]

        # Since nuScenes sampled at 2Hz
        times = np.linspace(
            0.0,
            0.5 * (len(sample_annotations) - 1),
            len(sample_annotations),
        )

        # Now link from instance IDs to relevant instance (entity) data
        instance_token_to_data: Dict[str, NuScenesInstanceData] = {}
        for sample, time in zip(sample_annotations, times):
            for annotation in sample:
                instance_token = annotation["instance_token"]
                if instance_token not in instance_token_to_data.keys():
                    instance_token_to_data[instance_token] = NuScenesInstanceData(
                        annotation["category_name"]
                    )
                instance_token_to_data[instance_token].trajectory.append(
                    annotation["translation"]
                )
                instance_token_to_data[instance_token].times.append(time)
                instance_token_to_data[instance_token].sizes.append(
                    annotation["size"]
                )
                instance_token_to_data[instance_token].rotations.append(
                    annotation["rotation"]
                )

        map_name = self.predict_helper.get_map_name_from_sample_token(
            first_sample_token
        )

        (entities, road_network,) = self._convert_to_entities_road_network(
            instance_token_to_data, map_name, ego_instance_token=ego_instance_token
        )

        scenario = Scenario(
            entities,
            name=scene_token,
            road_network=road_network,
        )
        return scenario

    def _convert_to_entities_road_network(
        self,
        instance_token_to_data,
        map_name,
        ego_instance_token: Optional[str] = None,
    ) -> Tuple[list[Entity], RoadNetwork]:
        if ego_instance_token is not None:
            if ego_instance_token not in instance_token_to_data.keys():
                raise KeyError("Ego instance token not found in scene.")
        else:
            potential_ego_instance_tokens = [
                i
                for i, d in instance_token_to_data.items()
                if "vehicle.car" in d.category_name
            ]
            if len(potential_ego_instance_tokens) == 0:
                raise ValueError("No potential ego vehicles in scene (no cars).")
            ego_instance_token = choice(potential_ego_instance_tokens)
            print(f"Chose ego instance token {ego_instance_token}")

        entities: Entity = []

        instance_tokens = set(instance_token_to_data.keys()) - set(
            [ego_instance_token]
        )
        instance_tokens = [ego_instance_token] + list(instance_tokens)
        for instance_token in instance_tokens:
            instance_data = instance_token_to_data[instance_token]
            entity_type = (
                "Pedestrian"
                if instance_data.category_name.split(".")[0]
                in {"human", "pedestrian"}
                else "Vehicle"
            )
            entity_category = "_".join(instance_data.category_name.split(".")[1:])
            # The scenario gym works in a slightly different way from the nuScenes
            # dataset. Bounding boxes are set at the catalog level and are fixed in
            # time.
            bounding_box = BoundingBox(
                *np.array(instance_data.sizes).mean(axis=0)[:2], 0.0, 0.0
            )
            setattr(
                Catalogs,
                instance_token,
                CatalogEntry(
                    Catalogs.nuScenes_catalog,
                    instance_data.category_name,
                    entity_category,
                    entity_type,
                    bounding_box,
                    {},
                    [],
                ),
            )

            rotations = Rotation.from_quat(instance_data.rotations).as_euler("xyz")[
                :, 0
            ]

            sg_rotations = np.pi * np.ones_like(rotations) - rotations

            traj_data = np.vstack(
                [
                    np.array(instance_data.times),
                    np.array(instance_data.trajectory).T[:2],
                    sg_rotations,
                ]
            ).T

            trajectory = Trajectory(traj_data, fields=["t", "x", "y", "h"])

            entity_ref = (
                f"entity_{instance_token}"
                if instance_token != ego_instance_token
                else "ego"
            )
            entity = Entity(getattr(Catalogs, instance_token), ref=entity_ref)
            entity.trajectory = trajectory
            entities.append(entity)

        all_trajectory_data = np.vstack(
            [np.array(d.trajectory) for d in instance_token_to_data.values()]
        )

        x_range, y_range, _ = np.ptp(all_trajectory_data, axis=0)
        radius = max([x_range, y_range]) * self.map_radius_multiplier

        centre_coordinate = np.mean(all_trajectory_data, axis=0)[:2]

        road_network = self._convert_nuScenes_map_to_road_network(
            map_name, centre_coordinate, radius
        )

        return entities, road_network



# Implement Mobileye's 5 rules of Responsibility-Sensitive Safety (RSS) as
# a metric compatible with driving gym.
# The five rules are
#   1)  Safe longitudinal distance to vehicles in the same lane
#   2)  Safe lateral distance to vehicles when merging lane
#   3)  Forfeiting right of way at junctions when another road user undermines
#       correct priority
#   4)  Additional caution in areas of occlusion
#   5)  Avoid a collision by any means without jeopardising safety of other
#       road users
from enum import Enum
from typing import Dict, List

from scenario_gym.metrics import Metric
from scenario_gym.road_network import road_network
from scenario_gym.state import State

from .callback import RSSDistances


class Rules(Enum):
    """Enumerate the five rules."""

    safe_longitudinal = 0
    safe_lateral = 1


class RSSBehaviourDetection:
    """
    Find behaviours corresponding to the five RSS rules.

    Instantiates class based on current timestep state and
    parameters. Calls each of the behaviour methods, one per
    RSS rule. These methods are lightweight, with most computation
    performed in the rss callback. Each method returns a bool
    value corresponding to whether or not the rule is obeyed at each
    timestep -- if the rule has not already been failed.
    """

    # Initialise and call the behaviour functions
    def __init__(
        self,
        metrics: Dict,
        ego: Dict,
        entities: List[Dict],
        safe_distances: List[List[float]],
        road_network: road_network,
        dt: float,
        intersect: List[List[str]],
        collisions,
    ):
        self.metrics = metrics
        self.ego = ego
        self.entities = entities
        self.safe_distances = safe_distances
        self.road_network = road_network
        self.dt = dt
        self.intersect = intersect
        self.collisions = collisions

    def __call__(self):
        """Call behaviour methods for current timestep."""
        outcomes = {}
        for rule in Rules:
            outcome = getattr(self, rule.name)
            outcomes[rule.name] = outcome()
        intersect = self.intersect
        return outcomes, intersect

    # Behaviour functions
    def safe_longitudinal(self) -> bool:
        """
        Rule 1: ego at a safe longitudinal distance.

        Returns True if longitudinal distance is less than the minimum
        safe distance when the lateral distance is also unsafe, with the safe
        longitudinal distance crossed last.
        """
        if not self.metrics["safe_longitudinal"]:
            # Already found
            return True
        # Check if any entities have been flagged as being unsafe longitudinally.
        for entity_record in self.intersect.values():
            if "unsafe_longitudinal" in entity_record:
                return False
        return True

    def safe_lateral(self) -> bool:
        """
        Rule 2: ego at a safe lateral distance.

        Returns True if lateral distance is less than the minimum
        safe distance when the longitudinal distance is also unsafe, with the
        safe lateral distance crossed last.
        """
        if not self.metrics["safe_lateral"]:
            # Already found
            return True
        # Check if any entities have been flagged as being unsafe laterally.
        for entity_record in self.intersect.values():
            if "unsafe_lateral" in entity_record:
                return False
        return True


class RSS(Metric):
    """
    Determine if the ego follow the 5 rules of RSS.

    _reset() resets state with each rule set to True by default

    _step() calls RSSBehaviourDetection to check if the ego is
    obeying the RSS rules at the current timestep. rss/callback
    is responsible for bulk of the computation and is called prior
    to the metric call at each timestep.

    get_state() returns dictionary of bools, one per rule
    --  True: the rule is obeyed at every timestep
    --  False: the rule is disobeyed in at least one timestep
        (may not be fault of ego - separate metric for blame)
    """

    required_callbacks = [RSSDistances]

    def _reset(self, state: State) -> None:
        """Reset behaviour."""
        self.rss_callback = self.callbacks[0]
        self.behaviour = None
        self.ego = state.scenario.ego
        self.metrics_ = {rule.name: True for rule in Rules}

    def _step(self, state: State) -> None:
        """Update the metric to find behaviour at the point of interest."""
        if state.t == 0.0:
            # Require at least two poses to calculate velocity
            return

        ego, entities, safe_distances, intersect = (
            self.rss_callback.ego_params,
            self.rss_callback.entity_params,
            self.rss_callback.safe_distances,
            self.rss_callback.intersect,
        )
        rules = RSSBehaviourDetection(
            metrics=self.metrics_,
            ego=ego,
            entities=entities,
            safe_distances=safe_distances,
            road_network=state.scenario.road_network,
            dt=state.dt,
            intersect=self.rss_callback.intersect,
            collisions=state.collisions(),
        )

        outcomes, intersect = rules()
        self.intersect = intersect
        for rule, outcome in outcomes.items():
            if outcome is False:
                self.metrics_[rule] = outcome

    def get_state(self) -> Dict[str, bool]:
        """Return state."""
        return self.metrics_



import numpy as np
import pytest as pt

from scenario_gym.scenario.actions import UpdateStateVariableAction
from scenario_gym.scenario_gym import ScenarioGym
from scenario_gym.state import State
from scenario_gym.xosc_interface import import_scenario


@pt.fixture
def scenario_path(all_scenarios):
    """Get a path for a scenario to test."""
    return all_scenarios["3e39a079-5653-440c-bcbe-24dc9f6bf0e6"]


@pt.fixture
def t0_scenario(all_scenarios):
    """Get a path for a scenario to test."""
    pth = all_scenarios["3e39a079-5653-440c-bcbe-24dc9f6bf0e6"]
    return import_scenario(pth).reset_start()


@pt.fixture
def scenario(scenario_path):
    """Get a scenario to test."""
    s = import_scenario(scenario_path)
    action = UpdateStateVariableAction(3.0, "TestAction", "ego", {"var": 1.0})
    s.add_action(action, inplace=True)
    return s


def test_poses(scenario):
    """Test the basic pose data recoreded in the gym state."""
    gym = ScenarioGym(timestep=0.1)
    gym.set_scenario(scenario)

    assert gym.state.t == gym.state.scenario.ego.trajectory.min_t
    assert gym.state.poses
    assert all(
        (
            np.allclose(
                v, (gym.state.poses[e] - gym.state.prev_poses[e]) / gym.state.dt
            )
            for e, v in gym.state.velocities.items()
        )
    ), "Velocities not correct."
    print(gym.state.recorded_poses())
    assert all(
        len(poses) == 2
        for e, poses in gym.state.recorded_poses().items()
        if e in gym.state.poses
    ), "Wrong number of recorded poses."

    gym.step()

    assert gym.state.poses
    assert all(
        (
            np.allclose(
                v, (gym.state.poses[e] - gym.state.prev_poses[e]) / gym.state.dt
            )
            for e, v in gym.state.velocities.items()
        )
    ), "Velocities not correct."
    assert all(
        len(poses) == 3
        for e, poses in gym.state.recorded_poses().items()
        if e in gym.state.poses
    ), "Wrong number of recorded poses."


def test_state_info(scenario):
    """Test running a scenario and getting data from the state."""
    gym = ScenarioGym(timestep=0.1)
    gym.set_scenario(scenario)

    for _ in range(50):
        gym.step()

    assert len(gym.state.poses) >= 2, "Not enough entities in the state."
    e = gym.state.scenario.entities[0]
    pose = gym.state.poses[e]
    distances = [
        np.linalg.norm(pose_[:3] - pose[:3])
        for e_, pose_ in gym.state.poses.items()
        if e_ != e
    ]

    assert (
        len(gym.state.get_entities_in_radius(*pose[:2], np.min(distances) - 0.1))
        == 1
    ) and (
        len(gym.state.get_entities_in_radius(*pose[:2], np.max(distances) + 1))
        == 1 + len(distances)
    ), "Incorrect entities returned."

    names, _ = gym.state.get_road_info_at_entity(e)
    assert "Road" in names, "Entity is on the road."


def test_step(t0_scenario):
    """Test the basic pose data recorded in the gym state."""
    gym = ScenarioGym(timestep=0.1)
    gym.set_scenario(t0_scenario)
    (ego, hazard) = gym.state.scenario.entities[:2]

    current = gym.state.poses.copy()
    next_poses = current.copy()
    next_poses[ego] = np.random.randn(6)
    next_poses[hazard] = np.random.randn(6)

    gym.state.next_t = 1.0
    gym.state.step(next_poses)

    assert np.allclose(gym.state.poses[ego], next_poses[ego])
    assert np.allclose(gym.state.poses[hazard], next_poses[hazard])
    assert gym.state.t == 1.0


def test_step_with_vanishing(t0_scenario):
    """Test updating the state with vanishing entities."""
    scenario = t0_scenario.reset_start(t0_scenario.entities[1])

    gym = ScenarioGym(timestep=0.1)
    gym.set_scenario(scenario)
    (ego, hazard) = gym.state.scenario.entities[:2]

    current = gym.state.poses.copy()
    current.pop(hazard)

    gym.state.next_t = 1.0
    gym.state.step(current)

    assert np.allclose(gym.state.poses[ego], current[ego])
    assert hazard not in gym.state.poses, "Vanishing entity not removed."
    assert hazard not in gym.state.prev_poses, "Vanishing entity not removed."
    assert hazard not in gym.state.velocities, "Vanishing entity not removed."
    assert hazard in gym.state.distances, "Vanishing entity should stay here."
    assert (
        hazard in gym.state.recorded_poses()
    ), "Vanishing entity should stay here."
    assert gym.state.t == 1.0

    gym.state.next_t = 2.0
    current[hazard] = hazard.trajectory.position_at_t(2.0)
    haz_prev = hazard.trajectory.position_at_t(1.0)
    haz_v = (current[hazard] - haz_prev) / gym.state.dt
    gym.state.step(current)

    assert hazard in gym.state.poses, "Vanishing entity not returned."
    assert hazard in gym.state.prev_poses, "Vanishing entity not returned."
    assert hazard in gym.state.velocities, "Vanishing entity not returned."
    assert np.allclose(
        gym.state.poses[hazard], current[hazard]
    ), "Vanishing entity has wrong pose."
    assert np.allclose(
        gym.state.prev_poses[hazard], haz_prev
    ), "Vanishing entity has wrong prev pose."
    assert np.allclose(
        gym.state.velocities[hazard], haz_v
    ), "Vanishing entity has wrong velocity."


def test_reset(scenario_path):
    """Test resetting the scenario with vanishing entities."""
    scenario = import_scenario(scenario_path)
    state = State(scenario)

    n = sum(1 for e in scenario.entities if e.trajectory.min_t <= 0.0)
    state.reset(-1.0, 0.0)
    assert len(state.poses) == n, "Wrong number of entities."
    assert len(state.prev_poses) == n, "Wrong number of entities."

    n = sum(1 for e in scenario.entities if e.trajectory.max_t >= 100.0)
    state.reset(-1.0, 100.0)
    assert len(state.poses) == n, "Wrong number of entities."
    assert len(state.poses) == n, "Wrong number of entities."


def test_reset_persist(scenario_path):
    """Test resetting the scenario with vanishing entities."""
    scenario = import_scenario(scenario_path)
    state = State(scenario, persist=True)

    n = len(scenario.entities)
    state.reset(-1.0, 0.0)
    assert len(state.poses) == n, "Wrong number of entities."
    assert len(state.prev_poses) == n, "Wrong number of entities."

    state.reset(-1.0, 100.0)
    assert len(state.poses) == n, "Wrong number of entities."
    assert len(state.poses) == n, "Wrong number of entities."


def test_state_actions(scenario):
    """Test simulation a scenario with actions."""
    gym = ScenarioGym(timestep=0.1)
    gym.set_scenario(scenario)
    assert not gym.state.entity_state[
        scenario.entities[0]
    ], "No actions should be applied."
    gym.rollout()
    assert (
        gym.state.entity_state[scenario.entities[0]]["var"] == 1.0
    ), "Action not applied."


def test_to_scenario(all_scenarios) -> None:
    """
    Rollout a single scenario and write to a new scenario.

    Output the xosc then load it again and rollout the
    recorded version.
    """
    scenario_path = all_scenarios["a5e43fe4-646a-49ba-82ce-5f0063776566"]

    # note we must reset start so that the recorded version will match the input
    scenario = import_scenario(scenario_path).reset_start()

    # rollout
    gym = ScenarioGym()
    gym.set_scenario(scenario)
    gym.rollout()

    poses = gym.state.recorded_poses()[scenario.entities[0]]
    assert np.unique(poses, axis=0).shape[0] == poses.shape[0]
    new_scenario = gym.state.to_scenario()

    ego = new_scenario.entities[0]
    assert len(ego.trajectory.t) == ego.trajectory.data.shape[0]

    # reload and test
    traj1 = scenario.entities[0].trajectory
    n_entities = len(scenario.entities)
    n_stationary = sum(1 for t in scenario.trajectories.values() if len(t) == 1)

    traj2 = new_scenario.entities[0].trajectory
    assert (
        len(new_scenario.entities) == n_entities
    ), "New scenario has a different number of entities."
    assert all(
        (
            isinstance(entity, type(old_entity))
            for entity, old_entity in zip(scenario.entities, new_scenario.entities)
        )
    ), "Entities are not the same type."
    assert n_stationary == sum(
        1 for t in new_scenario.trajectories.values() if len(t) == 1
    ), "New scenario has a different number of stationary entities."
    assert all(
        [
            np.allclose(traj1.position_at_t(0.0), traj2.position_at_t(0.0)),
            np.allclose(traj1.position_at_t(5.0), traj2.position_at_t(5.0)),
            np.allclose(traj1.position_at_t(10.0), traj2.position_at_t(10.0)),
        ]
    ), "Recorded and true trajectories differ."



from collections import OrderedDict
from dataclasses import dataclass
from typing import Any, Optional, Tuple, Type

import numpy as np

from scenario_gym.entity import Entity


@dataclass
class Observation:
    """Base class for an observation."""

    pass


@dataclass
class SingleEntityObservation(Observation):
    """Data for a single entity."""

    entity: Entity
    t: float
    next_t: float
    pose: np.ndarray
    velocity: np.ndarray
    distance_travelled: float
    recorded_poses: np.ndarray
    entity_state: Any


def combine_observations(
    *obs: Tuple[Type[Observation], ...],
    prefixes: Optional[Tuple[Optional[str], ...]] = None,
) -> Type[Observation]:
    """
    Create a class to combine multiple observations.

    The class will inherit the type annotations from the input classes and create
    a dataclass from them. If duplicate field names are found the `prefxies`
    argument can be used to create unique names by specifying a prexfix used
    for arguments from each observation. The created class will accept inputs
    from

    """
    if prefixes is not None and len(prefixes) != len(obs):
        raise ValueError

    annots = OrderedDict()
    maps = OrderedDict()
    for idx, ob in enumerate(obs):
        try:
            fields = ob.__dataclass_fields__
        except AttributeError as e:
            raise f"Obsevation {ob} not a dataclass." from e
        for f in fields.values():
            name = f.name
            if name in annots:
                if prefixes is None:
                    continue
                else:
                    pre = prefixes[idx]
                    name = f"{pre}_{name}"
                    if name in annots:
                        raise ValueError(
                            f"Prefix {pre} still leads to duplicate name for "
                            f"{name}."
                        )
            annots[name] = f.type
            maps[name] = (idx, name)

    @classmethod
    def from_obs(cls, *obs):
        """Create the class from observation instances."""
        args = []
        for (i, name) in maps.values():
            val = getattr(obs[i], name)
            args.append(val)
        return cls(*args)

    return dataclass(
        type(
            "CombinedObservation",
            (Observation,),
            {
                "__annotations__": annots,
                "from_obs": from_obs,
            },
        )
    )



import numpy as np

from scenario_gym.manager import ScenarioManager
from scenario_gym.metrics import Metric
from scenario_gym.scenario_gym import ScenarioGym


class EgoMaxSpeedMetric(Metric):
    """Measures the max speed achieved by the ego in the scenario."""

    max_speed: float

    def _reset(self, state):
        """Reset the metric."""
        self.ego = state.scenario.entities[0]
        self.max_speed = 0.0

    def _step(self, state):
        """Update the max speed."""
        self.max_speed = np.maximum(
            np.linalg.norm(state.velocities[self.ego][:3]),
            self.max_speed,
        )

    def get_state(self) -> float:
        """Return current max speed."""
        return self.max_speed


def test_manager():
    """Test that custom arguments can be supplied to the config."""
    manager = ScenarioManager()

    # change the timestep
    timestep = manager.timestep
    new_manager = ScenarioManager(timestep=timestep + 1)
    assert (
        new_manager.timestep == timestep + 1
    ), f"Should be {timestep+1} but got {new_manager.timestep}."

    # try adding a random argument
    new_manager = ScenarioManager(test_argument_12345="test")
    assert hasattr(new_manager, "test_argument_12345"), "Custom arg not assigned."
    assert new_manager.test_argument_12345 == "test", "Arg incorrectly assigned."


def test_add_metric(all_scenarios):
    """Test adding a metric to the scenario."""
    files = [
        all_scenarios[s]
        for s in (
            "3fee6507-fd24-432f-b781-ca5676c834ef",
            "41dac6fa-6f83-461e-a145-08692da5f3c7",
            "9c324146-be03-4d4e-8112-eaf36af15c17",
        )
    ]
    manager = ScenarioManager()
    manager.add_metric(EgoMaxSpeedMetric())
    output = manager.run_scenarios(files)
    assert len(output) == 3, "Should have returned results for 3 scenarios."

    gym = ScenarioGym()
    gym.load_scenario(files[0])
    gym.rollout()


def test_viewer_params(all_scenarios):
    """Test adding a metric to the scenario."""
    s = all_scenarios["3fee6507-fd24-432f-b781-ca5676c834ef"]
    manager = ScenarioManager(
        viewer_params={
            "render_layers": ["driveable_surface", "text"],
            "codec": "mp4v",
        }
    )
    gym = manager.make_gym()
    gym.load_scenario(s)
    gym.render()
    assert gym.viewer.render_layers == [
        "driveable_surface",
        "text",
    ], "Parameter not passed."
    assert gym.viewer.codec == "mp4v", "Parameter not passed."

    # repeat with incorrect argument
    manager = ScenarioManager(magnification=2)
    gym = manager.make_gym()
    gym.load_scenario(s)
    gym.render()
    assert gym.viewer.mag != 2, "Parameter should not be passed."



from abc import abstractmethod
from typing import Tuple

from scenario_gym.agent import Agent
from scenario_gym.pedestrian.observation import PedestrianObservation


class BehaviourParameters:
    """Parameters for the behaviour model."""

    max_speed_factor = 1.3

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class PedestrianBehaviour:
    """Base class for pedestrian behavior models."""

    def __init__(self, params: BehaviourParameters):
        self.params = params
        self.max_speed_factor = params.max_speed_factor

    def step(
        self,
        observation: PedestrianObservation,
        agent: Agent,
    ) -> Tuple[float, float]:
        """
        Return the new speed and heading according to behavior model.

        Parameters
        ----------
        observation : PedestrianObservation
            The observation for the pedestrian.

        agent : Agent
            The pedestrian agent.

        Returns
        -------
        speed : float

        heading : float

        """
        return self._step(observation, agent)

    @abstractmethod
    def _step(self, observation: PedestrianObservation, agent: Agent) -> Tuple:
        """Return the new speed and heading according to behavior model."""
        pass



from abc import ABC, abstractclassmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from lxml.etree import Element
from scenariogeneration import xosc

from scenario_gym.utils import ArgsKwargs, load_properties_from_xml


@dataclass(frozen=True)
class Catalog:
    """A catalog for catalog entries."""

    name: str
    group_name: str

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Load the catalog from a dictionary."""
        return cls(data["name"], data["group_name"])

    def to_dict(self) -> Dict[str, Any]:
        """Write the catalog to a dictionary."""
        return {"name": self.name, "group_name": self.group_name}


class CatalogObject(ABC):
    """
    Base class for objects loaded from catalogs.

    Subclasses should implement the `load_data_from_xml` class method wwhich takes
    the specific xml element that contains the data and returns the arguements
    and keyword arguments for the class constructor. It should not return the
    class itself since this way the methods can make use of loading methods from
    parent classes.

    The attribute xosc_namex can be set to the element names that the object
    represents. For example, if `xosc_names = ["Vehicle"]` then any elements with
    the tag `Vehicle` will be loaded by this entry. If not set then the class name
    will be used.
    """

    xosc_names: Optional[List[str]] = None

    @classmethod
    def from_xml(cls, element: Element, catalog: Optional[Catalog] = None):
        """Create the class from an xml element."""
        args, kwargs = cls.load_data_from_xml(element, catalog=catalog)
        return cls(*args, **kwargs)

    @abstractclassmethod
    def load_data_from_xml(
        cls,
        element: Element,
        catalog: Optional[Catalog] = None,
    ) -> ArgsKwargs:
        """Load the object from an xml element."""
        raise NotImplementedError

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """
        Create the object from a dictionary.

        Must be implemented to allow json serialization.
        """
        raise NotImplementedError

    def to_dict(self) -> Dict[str, Any]:
        """
        Write the object to a dictionary.

        Must be implemented to allow json serialization.
        """
        raise NotImplementedError

    def to_xosc(self) -> xosc.VersionBase:
        """Write the object to an xosc object."""
        raise NotImplementedError


@dataclass
class BoundingBox(CatalogObject):
    """A bounding box defined by its length, width and center."""

    width: float
    length: float
    center_x: float
    center_y: float

    @classmethod
    def load_data_from_xml(
        cls,
        element: Element,
        catalog: Optional[Catalog] = None,
    ) -> ArgsKwargs:
        """Load the bounding box data form an xml element."""
        if element.tag != "BoundingBox":
            raise TypeError(f"Expected BoundingBox element not {element.tag}.")
        bb_center = element.find("Center")
        bb_dimensions = element.find("Dimensions")
        return (
            float(bb_dimensions.attrib["width"]),
            float(bb_dimensions.attrib["length"]),
            float(bb_center.attrib["x"]),
            float(bb_center.attrib["y"]),
        ), {}

    @classmethod
    def from_dict(cls, data: Dict[str, float]):
        """Load the bounding box from a dictionary."""
        return cls(
            data["width"],
            data["length"],
            data["center_x"],
            data["center_y"],
        )

    def to_dict(self) -> Dict[str, float]:
        """Write the bounding box to a jsonable dictionary."""
        return {
            "width": self.width,
            "length": self.length,
            "center_x": self.center_x,
            "center_y": self.center_y,
        }

    def to_xosc(self) -> xosc.BoundingBox:
        """Write the bounding box to an xosc bounding box."""
        return xosc.BoundingBox(
            self.width,
            self.length,
            0.0,
            self.center_x,
            self.center_y,
            0.0,
        )


@dataclass
class CatalogEntry(CatalogObject):
    """
    A single catalog entry. Holds catalog information and a bounding box.

    Parameters
    ----------
    catalog : Optional[Catalog]
        The catalog from which the entry is loaded.

    catalog_entry : str
        The name of the specific entry.

    catalog_category : Optional[str]
        The category of the entry.

    catalog_type : str
        The catalog type e.g Vehicle or Pedestrian.

    bounding_box : BoundingBox
        The bounding box of the entry.

    properties : Dict[str, Union[float, str]]
        Any properties associated with the element.

    files: List[str]
        A list of filepaths for external files.

    """

    catalog: Optional[Catalog]
    catalog_entry: str
    catalog_category: Optional[str]
    catalog_type: str
    bounding_box: BoundingBox
    properties: Dict[str, Union[float, str]]
    files: List[str]

    @classmethod
    def load_data_from_xml(
        cls,
        element: Element,
        catalog: Optional[Catalog] = None,
    ) -> ArgsKwargs:
        """Load the catalog entry from an xml element."""
        entry_name = element.attrib["name"]
        cname = element.tag.lower() + "Category"
        category = element.attrib[cname] if cname in element.attrib else None
        bb = element.find("BoundingBox")
        bb = BoundingBox.from_xml(bb, catalog=catalog)
        properties, files = load_properties_from_xml(element)
        return (
            catalog,
            entry_name,
            category,
            element.tag,
            bb,
            properties,
            files,
        ), {}

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Load the catalog entry from a dictionary."""
        catalog = data.get("catalog", None)
        if catalog is not None:
            catalog = Catalog.from_dict(catalog)
        return cls(
            catalog,
            data["catalog_entry"],
            data["catalog_category"],
            data["catalog_type"],
            BoundingBox.from_dict(data["bounding_box"]),
            data.get("properties", {}),
            data.get("files", []),
        )

    def to_dict(self) -> Dict[str, Any]:
        """Write the catalog entry to a dictionary."""
        return {
            "catalog": self.catalog.to_dict() if self.catalog else None,
            "catalog_entry": self.catalog_entry,
            "catalog_category": self.catalog_category,
            "catalog_type": self.catalog_type,
            "bounding_box": self.bounding_box.to_dict(),
            "properties": self.properties,
            "files": self.files,
        }

    def to_xosc(self) -> xosc.VersionBase:
        """Create an xosc entity object from the catalog entry."""
        obj = xosc.MiscObject(
            self.catalog_entry,
            1.0,
            getattr(
                xosc.MiscObjectCategory,
                self.catalog_category,
                xosc.MiscObjectCategory.none,
            ),
            self.catalog_category,
            self.bounding_box.to_xosc(),
        )
        for k, v in self.properties.items():
            obj.add_property(k, v)
        for f in self.files:
            obj.add_property_file(f)
        return obj



import pytest as pt

from scenario_gym.scenario.actions import UserDefinedAction
from scenario_gym.xosc_interface import import_scenario


@pt.fixture(scope="module")
def scenario(all_scenarios):
    """Get a scenario with actions."""
    path = all_scenarios["1518e754-318f-4847-8a30-2dce552b4504"]
    return import_scenario(path)


def test_user_actions(scenario):
    """Test getting user actions."""
    assert scenario.actions, "There should be actions."

    act = scenario.actions[0]
    assert isinstance(act, UserDefinedAction), "Incorrect action type."
    assert act.action_class == "CustomCommandAction", "Incorrect action class."
    assert act.entity_ref == "ego", "Incorrect entity ref."
    assert act.action_variables == {"type": "turn_right"}, "Incorrect variables."
    assert act.t == 5.0, "Incorrect time."



from abc import ABC, abstractmethod
from typing import Optional

from scenario_gym.entity import Entity
from scenario_gym.observation import Observation
from scenario_gym.state import State


class Sensor(ABC):
    """
    Produce an observation for a given entity from the global state.

    All sensors should implement the `_reset` and `_step` method to produce
    observations which should contain all information that an agent may use to
    select actions. The `_reset` method should produce an initial observation for
    the agent. The `_step` method should produce the observation to be used at each
    timestep.
    """

    def __init__(self, entity: Entity):
        """Init the sensor."""
        self.entity = entity
        self.initial_observation: Optional[Observation] = None
        self._last_observation: Optional[Observation] = None

    def reset(self, state: State) -> Observation:
        """Reset the sensor at the start of the scenario."""
        self._last_observation = None
        self.initial_observation = self._reset(state)
        return self.initial_observation

    def step(self, state: State) -> Observation:
        """Produce the observation from the global state."""
        self.last_observation = self._step(state)
        return self.last_observation

    @abstractmethod
    def _reset(self, state: State) -> Observation:
        """Reset the sensor and return an initial observation."""
        raise NotImplementedError

    @abstractmethod
    def _step(self, state: State) -> Observation:
        """Produce an observation from the global state."""
        raise NotImplementedError

    @property
    def last_observation(self) -> Optional[Observation]:
        """Get the previous observation."""
        return self._last_observation

    @last_observation.setter
    def last_observation(self, obs: Optional[Observation]) -> None:
        self._last_observation = obs



from typing import Tuple

import numpy as np

from scenario_gym.agent import Agent
from scenario_gym.pedestrian.behaviour import (
    BehaviourParameters,
    PedestrianBehaviour,
)
from scenario_gym.pedestrian.observation import PedestrianObservation


class RandomWalkParameters(BehaviourParameters):
    """Parameters for the random walk model."""

    bias_lon = 0.0
    bias_lat = 0.0
    std_lon = 0.000002
    std_lat = 0.0000001


class RandomWalk(PedestrianBehaviour):
    """Random walk model."""

    def __init__(self, params: RandomWalkParameters):
        super().__init__(params)
        self.bias_lon = params.bias_lon
        self.bias_lat = params.bias_lat
        self.std_lon = params.std_lon
        self.std_lat = params.std_lat

    def _step(
        self,
        observation: PedestrianObservation,
        agent: Agent,
    ) -> Tuple[float, float]:
        """Return the new speed and heading according to the random walk model."""
        speed_rand = np.random.normal(
            agent.speed_desired + self.bias_lon, self.std_lon
        )
        goal = agent.route[agent.goal_idx] - observation.pose[[0, 1]]
        heading = np.arctan2(goal[1], goal[0])  # angle to goal point
        heading_rand = np.random.normal(heading + self.bias_lat, self.std_lat)
        return speed_rand, heading_rand



from __future__ import annotations

import json
import warnings
from contextlib import suppress
from copy import copy
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type

import matplotlib.pyplot as plt
import numpy as np

from scenario_gym.entity import Entity, MiscObject, Pedestrian, Vehicle
from scenario_gym.road_network import RoadNetwork
from scenario_gym.scenario.actions import ScenarioAction, UpdateStateVariableAction
from scenario_gym.trajectory import Trajectory
from scenario_gym.utils import cached_property


class Scenario:
    """
    The scenario_gym representation of a scenario.

    A scenario consists of a set of entities and a road network. The entities have
    trajectories and catalog entries and may have additional entity specific
    properties. The scenario also may have a list of actions which can specify
    events that occur. E.g. a traffic light changing color.
    """

    def __init__(
        self,
        entities: List[Entity],
        name: Optional[str] = None,
        road_network: Optional[RoadNetwork] = None,
        actions: Optional[List[ScenarioAction]] = None,
        properties: Optional[Dict[Any, Any]] = None,
    ):
        self._entities = entities
        self._ref_to_entity: Dict[str, Entity] = {e.ref: e for e in entities}

        self.name = name
        self.road_network = road_network
        self.actions = actions if actions is not None else []
        self.properties = properties if properties is not None else {}

        self._vehicles: Optional[List[Entity]] = None
        self._pedestrians: Optional[List[Entity]] = None

    @property
    def entities(self) -> List[Entity]:
        """Get the entities in the scenario."""
        return self._entities

    @property
    def ego(self) -> Entity:
        """
        Get the ego entity.

        The ego entity is defined as the entity with the ref "ego" or the first
        entity if no entity has the ref "ego".
        """
        ego = self.entity_by_name("ego")
        if ego is not None:
            return ego
        return self.entities[0]

    @property
    def vehicles(self) -> List[Entity]:
        """Get the entities that have vehicle catalogs."""
        if self._vehicles is None:
            self._vehicles = [e for e in self.entities if isinstance(e, Vehicle)]
        return self._vehicles

    @property
    def pedestrians(self) -> List[Entity]:
        """Get the entities that have pedestrian catalogs."""
        if self._pedestrians is None:
            self._pedestrians = [
                e for e in self.entities if isinstance(e, Pedestrian)
            ]
        return self._pedestrians

    @property
    def trajectories(self) -> Dict[str, Trajectory]:
        """Return a dictionary mapping entity references to the trajectory."""
        return {e.ref: e.trajectory for e in self.entities}

    @cached_property
    def length(self) -> float:
        """Return the length of the scenario in seconds."""
        return max([t.max_t for t in self.trajectories.values()])

    def entity_by_name(self, e_ref: str) -> Optional[Entity]:
        """Return an entity given a unique reference."""
        with suppress(KeyError):
            return self._ref_to_entity[e_ref]

    def __copy__(self) -> Scenario:
        """Create a copy of a scenario without copying the road network."""
        return self.__class__(
            name=f"Copy of {self.name}" if self.name is not None else None,
            road_network=self.road_network,
            actions=[a.copy() for a in self.actions],
            entities=[e.copy() for e in self.entities],
            properties=self.properties,
        )

    def copy(self) -> Scenario:
        """Create a copy of the scenario."""
        return copy(self)

    def add_entity(self, e: Entity, inplace: bool = False) -> Scenario:
        """Create a new scenario with the entity added."""
        if e.ref in self._ref_to_entity:
            i = 0
            while True:
                new_ref = f"{e.ref}_{i}"
                if new_ref not in self._ref_to_entity:
                    break
                i += 1
            old_ref = e.ref
            e.ref = new_ref
            warnings.warn(
                f"An entity with ref {old_ref} exists. Adding with ref {new_ref}."
            )
        scenario = self.copy() if not inplace else self
        scenario._entities.append(e)
        scenario._ref_to_entity[e.ref] = e
        scenario._vehicles = None
        scenario._pedestrians = None
        return scenario

    def remove_entity(self, e: Entity, inplace: bool = False) -> Scenario:
        """Create a new scenario with the entity added."""
        idx = self._entities.index(e)
        scenario = self.copy() if not inplace else self
        scenario._entities.pop(idx)
        scenario._ref_to_entity.pop(e.ref)
        scenario._vehicles = None
        scenario._pedestrians = None
        return scenario

    def make_ego(self, e: Entity, inplace: bool = False) -> Scenario:
        """Set e to the ego entity."""
        try:
            idx = self._entities.index(e)
        except ValueError:
            idx = None
        scenario = self.copy() if not inplace else self

        if idx is not None:
            e = scenario._entities.pop(idx)
        else:
            scenario._ref_to_entitiy[e.ref] = e
        scenario._entities.insert(0, e)
        scenario.vehicles = None
        scenario.pedestrians = None
        return scenario

    def add_action(self, action: ScenarioAction, inplace: bool = False) -> Scenario:
        """Add an action to the scenario."""
        scenario = self.copy() if not inplace else self
        scenario.actions.append(action)
        return scenario

    def translate(self, x: np.ndarray, inplace: bool = False) -> Scenario:
        """Return a new scenario with all entities translated."""
        scenario = self.copy() if not inplace else self
        for e in scenario.entities:
            e.trajectory = e.trajectory.translate(x)

        actions = []
        for a in scenario.actions:
            actions.append(a.translate(x, inplace=inplace))
        scenario.actions = actions

        return scenario

    def reset_start(self, entity: Optional[Entity] = None) -> Scenario:
        """Reset the start time to the start of an entity's trajectory."""
        if entity is None:
            entity = self.ego
        start_time = entity.trajectory.min_t
        return self.translate(np.array([-start_time, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]))

    @classmethod
    def from_dict(
        cls,
        data: Dict[str, Any],
        e_classes: Tuple[Type[Entity]] = (Vehicle, Pedestrian, Entity),
        a_classes: Tuple[Type[ScenarioAction]] = (UpdateStateVariableAction,),
    ):
        """Load the scenario from a dictionary."""
        entities = []
        for e_data in data["entities"]:
            for Ent in e_classes:
                if Ent.__name__ == e_data.get("entity_class", Entity):
                    break
            entities.append(Ent.from_dict(e_data))

        road_network = data.get("road_network")
        if road_network is not None:
            if road_network.get("path") is not None:
                path = Path(road_network["path"])
                if path.exists():
                    road_network = RoadNetwork.create_from_file(path)
                elif road_network.get("name") is not None:
                    road_network = RoadNetwork(name=road_network["name"])
                else:
                    road_network = None
            else:
                road_network = RoadNetwork.create_from_dict(road_network)

        actions = []
        for a_data in data.get("actions", ()):
            for Act in a_classes:
                if Act.__name__ == a_data.get(
                    "action_class", "UpdateStateVariableAction"
                ):
                    break
            actions.append(Act.from_dict(a_data))

        return cls(
            entities,
            name=data.get("name"),
            road_network=road_network,
            actions=actions,
            properties=data.get("properties", {}),
        )

    def to_dict(
        self,
        road_network_path: Optional[str] = "../Road_Networks",
    ) -> Dict[str, Any]:
        """Write the scenario to a dictionary."""
        if self.road_network is None:
            road_network = None
        elif road_network_path is not None:
            if not Path(road_network_path).is_file():
                road_network_path = str(
                    Path(
                        road_network_path,
                        f"{self.road_network.name}.json",
                    )
                )
            road_network = {
                "path": road_network_path,
                "name": self.road_network.name,
            }
        else:
            road_network = self.road_network.to_dict()
        return {
            "entities": [e.to_dict() for e in self.entities],
            "name": self.name,
            "actions": [act.to_dict() for act in self.actions],
            "road_network": road_network,
            "properties": self.properties,
        }

    @classmethod
    def from_json(
        cls,
        path: str,
        road_network_dir: Optional[str] = None,
        e_classes: Tuple[Type[Entity]] = (Vehicle, Pedestrian, Entity),
        a_classes: Tuple[Type[ScenarioAction]] = (UpdateStateVariableAction,),
    ):
        """
        Load the scenario from a json file.

        Parameters
        ----------
        path : str
            The path to the json file.

        road_network_dir : str, optional
            The directory to search for the road network file. If the road network
            path in the json is absolute then this will be ignored. Otherwise the
            path used will be (road_network_dir / road network path) if the
            road_network_dir is relative otherwise it will be the `directory of
            path / road_network_dir / road network path`.

        e_classes : Tuple[Type[Entity]], optional
            The classes to use for loading entities.

        a_classes : Tuple[Type[ScenarioAction]], optional
            The classes to use for loading actions.

        """
        with open(path, "r") as f:
            data = json.load(f)
        if data.get("road_network", {}).get("path") is not None:
            rn_path = Path(data["road_network"]["path"])
            if not rn_path.is_absolute():
                if road_network_dir is None:
                    rn_path = Path(path).parent / rn_path
                elif Path(road_network_dir).is_absolute():
                    rn_path = Path(road_network_dir) / rn_path
                else:
                    rn_path = Path(
                        Path(path).parent,
                        road_network_dir,
                        rn_path,
                    )
                print(rn_path)
                data["road_network"]["path"] = str(rn_path)
        return cls.from_dict(
            data,
            e_classes=e_classes,
            a_classes=a_classes,
        )

    def to_json(
        self, path, road_network_path: Optional[str] = "../Road_Networks"
    ) -> None:
        """Write the scenario to a json file."""
        data = self.to_dict(road_network_path=road_network_path)
        with open(path, "w") as f:
            json.dump(data, f)

    def describe(self) -> None:
        """Generate a text overview of the scenario."""
        rn = self.road_network.name if self.road_network is not None else "None"
        name = (
            self.name.replace(".xosc", "") if self.name is not None else "scenario"
        )
        title = f"Scenario: {name}"
        under_header = "=" * len(title)

        entity_header = "Entity".ljust(10) + "Type".ljust(10) + "Cateogry".ljust(10)
        entities = ""
        for e in self.entities:
            entities += (
                f"{e.ref}".ljust(10)
                + f"{e.type}".ljust(10)
                + f"{e.catalog_entry.catalog_category}".ljust(10)
                + "\n"
            )

        print(
            f"""
{title}
{under_header}
Road network: {rn}
Number of entities: {len(self.entities)}
Total duration: {self.length:.4}s

Entities
--------
{entity_header}
{entities}
"""
        )

    def plot(self, figsize: Tuple[int, int] = (10, 10), show: bool = True) -> None:
        """
        Visualise the scenario.

        Parameters
        ----------
        figsize : Tuple[int, int]
            The figure size.

        show : bool
            If set to False will not call `plt.show` so the figure can be modified
            or saved.

        """
        name = self.name if self.name is not None else "Scenario"
        plt.figure(figsize=figsize)
        if self.road_network is not None:
            for geom in self.road_network.driveable_surface.geoms:
                plt.fill(*geom.exterior.xy, c="gray", alpha=0.25)
                for i in geom.interiors:
                    plt.fill(*i.xy, c="white")
            for r in self.road_network.roads:
                plt.plot(*r.center.xy, c="white")
        for i, e in enumerate(self.entities):
            if i == 0:
                c = "r"
            elif isinstance(e, Pedestrian):
                c = "g"
            elif isinstance(e, MiscObject):
                c = "gray"
            else:
                c = "b"
            plt.plot(*e.trajectory.data[:, [1, 2]].T, c=c, label=e.ref)
            plt.plot(*e.trajectory.data[0, [1, 2]].T, c=c, marker="o")
        data = np.vstack([e.trajectory.data[:, [1, 2]] for e in self.entities])
        b_min, b_max = data.min(0), data.max(0)
        plt.axis("equal")
        plt.xlim(b_min[0] - 10.0, b_max[0] + 10.0)
        plt.ylim(b_min[1] - 10.0, b_max[1] + 10.0)
        plt.legend()
        plt.title(name)
        if show:
            plt.show()



import json
from contextlib import suppress
from functools import _lru_cache_wrapper, lru_cache, partial
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type, Union

import numpy as np
from pyxodr.road_objects.network import RoadNetwork as xodrRoadNetwork
from scipy.interpolate import LinearNDInterpolator, NearestNDInterpolator
from scipy.spatial import Delaunay
from shapely.geometry import MultiPolygon, Point, Polygon
from shapely.ops import unary_union

from scenario_gym.utils import ArrayLike, NDArray, cached_property

from .base import RoadGeometry, RoadObject
from .objects import (
    Building,
    Crossing,
    Intersection,
    Lane,
    LaneType,
    Pavement,
    Road,
)
from .xodr import xodr_to_sg_roads


class RoadNetwork:
    """
    A collection of roads, intersections, etc that form a road network.

    The road network implements layers that give different objects in
    the network. Default layers can be seein the object_names attribute.

    Any list objects that subclasses RoadObject or RoadGeometry can be passed
    as keywords to add custom objects to the road network.
    """

    _default_object_names: Dict[str, Type[RoadObject]] = {
        "roads": Road,
        "intersections": Intersection,
        "lanes": Lane,
        "pavements": Pavement,
        "crossings": Crossing,
        "buildings": Building,
    }

    @classmethod
    def create_from_file(cls, filepath: str):
        """
        Create the road network from a file.

        Parameters
        ----------
        filepath : str
            The path to the file.

        """
        path = Path(filepath).absolute()
        if not path.exists():
            raise FileNotFoundError(f"File not found at: {path}.")

        if path.suffix in (".json", ""):
            return cls.create_from_json(filepath)
        elif path.suffix == ".xodr":
            return cls.create_from_xodr(filepath)
        raise ValueError(f"Unknown file type: {path.suffix}.")

    @classmethod
    @lru_cache(maxsize=15)
    def create_from_json(cls, filepath: str):
        """
        Create the road network from a json file.

        Parameters
        ----------
        filepath : str
            The path to the json file.

        """
        with open(filepath) as f:
            data = json.load(f)
        return cls.create_from_dict(data, name=Path(filepath).stem)

    @classmethod
    @lru_cache(maxsize=15)
    def create_from_xodr(
        cls,
        filepath: str,
        resolution: float = 0.1,
        simplify_tolerance: float = 0.2,
    ):
        """
        Import a road network from an OpenDRIVE file.

        Will first parse the road network and then convert it to
        a scenario_gym road network. Every lane section of the file
        is converted to a road and each lane within the section is
        converted to a lane. Connectivity information is stored in
        the lanes. Any lane of type None is ignored.

        Parameters
        ----------
        filepath : str
            The filepath to the xodr file.

        resolution : float
            Resolution for importing the base OpenDRIVE file.

        simplify_tolerance : float
            Points per m for simplifying center and boundary lines.

        """
        path = Path(filepath).absolute()
        if not path.exists():
            raise FileNotFoundError(f"File not found at: {path}.")

        # parse OpenDRIVE file
        xodr_network = xodrRoadNetwork(
            str(path),
            resolution=resolution,
        )

        roads = xodr_to_sg_roads(
            xodr_network,
            simplify_tolerance,
        )

        return cls(roads=roads, name=path.stem)

    @classmethod
    def create_from_dict(cls, data: Dict, **kwargs):
        """
        Create a road network from a dictoinary of road data.

        The dictionary must have keys 'Roads' and 'Intersections' and
        optionally with keys for other road objects. Each of these must hold
        a list of dicts with the data for each object. These should hold their
        required fields e.g. Center, Boundary, successors, predecessors.
        """
        assert (
            "Roads" in data or "roads" in data
        ), "Json data must contain road information."
        assert (
            "Intersections" in data or "intersections" in data
        ), "Json data must contain intersection information."

        objects = {}
        for obj, obj_cls in cls._default_object_names.items():
            if obj in data:
                key = obj
            elif obj.capitalize() in data:
                key = obj.capitalize()
            else:
                continue
            objects[obj] = [obj_cls.from_dict(obj_data) for obj_data in data[key]]

        if "name" not in kwargs and "name" in data:
            kwargs["name"] = data["name"]

        return cls(**kwargs, **objects)

    def __init__(
        self,
        name: Optional[str] = None,
        **road_objects: Dict[str, List[RoadObject]],
    ):
        """
        Construct the road network.

        This takes lists of road objects as keywords. The keyword used determines
        how the objects will be stored. E.g. `roads=[...]` will define the `roads`
        attribute. This way custom road objects can be passed e.g. passing
        `road_markings=[...]` will mean a `road_markings` attribute is created.
        Every object in the list must be a subclass of `RoadObject`. `roads` and
        `intersections` must be passed even if they are empty lists.

        Parameters
        ----------
        name: Optional[str]
            Optional name for the road network.

        road_objects : Dict[str, List[RoadObject]]
            The road objects as keywords. `roads` and `intersections` must be
            passed.

        """
        self.name = name

        # cached elevation interpolation functions
        self._hull = None
        self._inside_fn = None
        self._outisde_fn = None

        self._lane_parents: Dict[Lane, Optional[Union[Road, Intersection]]] = {}

        self.object_names = self._default_object_names.copy()
        self.object_classes = {v: k for k, v in self.object_names.items()}
        all_object_names = list(
            set(self.object_names.keys())
            .union(road_objects.keys())
            .difference(["roads", "intersections"])
        )
        for object_name in ["roads", "intersections"] + all_object_names:
            objects = (
                road_objects[object_name] if object_name in road_objects else []
            )
            assert all(
                (isinstance(obj, RoadObject) for obj in objects)
            ), "Only lists of RoadObject subclasses should be provided"

            if object_name not in self.object_names:
                self.object_names[object_name] = (
                    objects[0].__class__ if objects else RoadObject
                )
            self.add_new_road_object(objects, object_name)

    def add_new_road_object(
        self, objs: Union[RoadObject, List[RoadObject]], obj_name: str
    ) -> None:
        """
        Add a new object type to the road network.

        This will add an attribute for the raw objects as a list as well
        as a public attribute if it does not already exist. It will also add
        an add_{obj_name} method to add new objects to the list.
        """
        if hasattr(self, f"_{obj_name}"):
            raise ValueError(
                f"Road network already has {obj_name}. Use self.add_{obj_name}."
            )
        setattr(self, f"_{obj_name}", objs)
        try:
            getattr(self, obj_name)
        except AttributeError:
            setattr(self, obj_name, objs)
        try:
            getattr(self, f"add_{obj_name}")
        except AttributeError:
            setattr(
                self,
                f"add_{obj_name}",
                partial(self._add_obj, obj_name=obj_name),
            )

    def _add_obj(self, objs: List[RoadObject], obj_name: Optional[str] = None):
        if obj_name is None:
            raise ValueError("Must provide obj_name")
        getattr(self, f"_{obj_name}").extend(
            objs if isinstance(objs, list) else [objs]
        )
        self.clear_cache()

    @cached_property
    def roads(self) -> List[Road]:
        """Get all roads in the road network."""
        return self._roads

    @cached_property
    def intersections(self) -> List[Intersection]:
        """Get all intersections in the road network."""
        return self._intersections

    @cached_property
    def lanes(self) -> List[Lane]:
        """Get all lanes in the road network."""
        return list(
            set(sum([x.lanes for x in self.roads + self.intersections], [])).union(
                self._lanes
            )
        )

    @cached_property
    def road_network_objects(self) -> List[RoadObject]:
        """Get all the road objects in the network."""
        return [
            obj for obj_name in self.object_names for obj in getattr(self, obj_name)
        ]

    @cached_property
    def road_network_geometries(self) -> List[RoadGeometry]:
        """Get all road geometries in the network."""
        geoms = []
        for obj_name, obj_class in self.object_names.items():
            if issubclass(obj_class, RoadGeometry):
                geoms.extend(getattr(self, obj_name))
        return geoms

    @cached_property
    def driveable_surface(self) -> MultiPolygon:
        """Get the union of boundaries of driveable geometries."""
        merged = unary_union(
            [g.boundary for g in self.road_network_geometries if g.driveable]
        )
        return MultiPolygon([merged]) if isinstance(merged, Polygon) else merged

    @cached_property
    def walkable_surface(self) -> MultiPolygon:
        """Get the union of boundaries of non-driveable geometries."""
        merged = unary_union(
            [g.boundary for g in self.road_network_geometries if g.walkable]
        )
        return MultiPolygon([merged]) if isinstance(merged, Polygon) else merged

    @cached_property
    def impenetrable_surface(self) -> MultiPolygon:
        """Get the union of all impenetrable geometries."""
        merged = unary_union(
            [g.boundary for g in self.road_network_geometries if g.impenetrable]
        )
        return MultiPolygon([merged]) if isinstance(merged, Polygon) else merged

    def object_by_id(self, i: str) -> RoadObject:
        """Get the object with the given id."""
        return self._object_by_id[i]

    @cached_property
    def _object_by_id(self) -> Dict[str, RoadObject]:
        """Return a dict indexing all objects by id."""
        return {x.id: x for x in self.road_network_objects}

    @cached_property
    def driveable_lanes(self) -> List[Lane]:
        """Get all driveable lanes in the network."""
        return [l for l in self.lanes if l.type is LaneType["driving"]]

    @cached_property
    def _lanes_by_id(self) -> Dict[str, Lane]:
        """Return a dict indexing all lanes by id."""
        return {l.id: l for l in self.lanes}

    def get_successor_lanes(self, l: Lane) -> List[Lane]:
        """Get lanes that succeed the given lane."""
        return [self._lanes_by_id[l_] for l_ in l.successors]

    def get_predecessor_lanes(self, l: Lane) -> List[Lane]:
        """Get lanes that predecess the given lane."""
        return [self._lanes_by_id[l_] for l_ in l.predecessors]

    def get_connecting_roads(self, i: Intersection) -> List[Road]:
        """Get roads that connect to the given intersection."""
        return [r for r in self.roads if r in i.connecting_roads]

    def get_intersections(self, r: Road) -> List[Intersection]:
        """Get intersections that connect to the given road."""
        return [i for i in self.intersections if r in i.connecting_roads]

    def get_lane_parent(self, l: Lane) -> Optional[Union[Road, Intersection]]:
        """Get the object that the lane belongs to."""
        if l not in self._lane_parents:
            for x in self.roads + self.intersections:
                if l in x.lanes:
                    self._lane_parents[l] = x
                    return x
            self._lane_parents[l] = None
        return self._lane_parents[l]

    def get_geometries_at_point(
        self,
        x: float,
        y: float,
    ) -> Tuple[List[str], List[RoadGeometry]]:
        """
        Get all geometries at a given xy point.

        TODO: Move to a spatial index for speed.

        Parameters
        ----------
        x : float
            The x-coordinate at the point.

        y : float
            The y-coordinate at the point.

        Returns
        -------
        Tuple[List[str], List[RoadObject]]
            A list of string identifiers of the geometry (e.g. Road, Lane)
            and the actual objects.

        """
        p = Point(x, y)

        names, geoms = [], []
        for x in self.road_network_geometries:
            if x.boundary.contains(p):
                names.append(x.__class__.__name__)
                geoms.append(x)
        return names, geoms

    def to_dict(self) -> Dict[str, List[Dict[str, Any]]]:
        """Return a dict representation of the road network."""
        data = {"name": self.name}
        for obj_name in self.object_names:
            data[obj_name] = [obj.to_dict() for obj in getattr(self, obj_name)]
        return data

    def to_json(self, filepath: str) -> None:
        """Save the road network to json file."""
        data = self.to_dict()
        with open(filepath, "w") as f:
            json.dump(data, f)

    def clear_cache(self) -> None:
        """Clear the cached properties and lru cache methods."""
        self._lane_parents.clear()
        self._hull = None
        self._inside_fn = None
        self._outisde_fn = None
        for method in dir(self.__class__):
            obj = getattr(self.__class__, method)
            if isinstance(obj, _lru_cache_wrapper):
                getattr(self, method).__func__.cache_clear()
            elif (
                isinstance(cached_property, type)
                and isinstance(obj, cached_property)
                and (method in self.__dict__)
            ):
                del self.__dict__[method]
            else:
                with suppress(AttributeError):
                    func = obj.__func__
                    if isinstance(func, _lru_cache_wrapper) and (
                        obj.__self__ is self
                    ):
                        func.cache_clear()

    def elevation_at_point(self, x: ArrayLike, y: ArrayLike) -> NDArray:
        """Estimate the elevation at (x, y) by interpolating."""
        x = np.array(x)
        y = np.array(y)
        if self._hull is None:
            self._interpolate_elevation()

        x_ndim, y_ndim = x.ndim, y.ndim
        if x_ndim not in (0, 1) or y_ndim not in (0, 1):
            raise ValueError("x and y must be 0 or 1 dimensional.")

        if x_ndim == 0:
            x = np.array([x])
        if y_ndim == 0:
            y = np.array([y])

        if x.shape[0] == 1 and y.shape[0] > 1:
            x = np.repeat(x, y.shape[0])
        elif y.shape[0] == 1 and x.shape[0] > 1:
            y = np.repeat(y, x.shape[0])

        xy = np.column_stack((x, y))

        inside = self._hull.find_simplex(xy) >= 0
        res = np.empty(xy.shape[0])
        if np.any(inside):
            zs_in = self._inside_fn(xy[inside])
            res[inside] = zs_in
        if np.any(~inside):
            zs_out = self._outside_fn(xy[~inside])
            res[~inside] = zs_out

        if x_ndim == y_ndim == 1:
            res = res.squeeze()
        return res

    def _interpolate_elevation(self) -> None:
        """Interpolate the elevation values of the geometries."""
        elevs = [
            geom.elevation
            for geom in self.road_network_geometries
            if geom.elevation is not None
        ]
        if not elevs:
            # this used to be a lambda function returning zeros but
            # that was not pickleable so interp2d is used with zero inputs
            elevation_values = np.array(
                [
                    [0, 1, 0],
                    [1, 0, 0],
                    [1, 1, 0],
                    [0, 0, 0],
                ]
            )
        else:
            elevation_values = np.concatenate(elevs, axis=0)

        if elevation_values.shape[0] > 5000:
            n = np.ceil(elevation_values.shape[0] / 5000)
            elevation_values = elevation_values[:: int(n)]

        self._hull = Delaunay(elevation_values[:, :2])
        self._inside_fn = LinearNDInterpolator(
            elevation_values[:, :2],
            elevation_values[:, 2],
        )
        self._outside_fn = NearestNDInterpolator(
            elevation_values[:, :2],
            elevation_values[:, 2],
        )



import numpy as np

from scenario_gym.state import State

from .base import Metric


class EgoAvgSpeed(Metric):
    """Measure the average speed of the ego."""

    name = "ego_avg_speed"

    def _reset(self, state: State) -> None:
        """Reset the average speed."""
        self.ego = state.scenario.ego
        self.ego_avg_speed = np.linalg.norm(state.velocities[self.ego][:3])
        self.t = 0.0

    def _step(self, state: State) -> None:
        """Update the average speed."""
        speed = np.linalg.norm(state.velocities[self.ego][:3])
        w = self.t / state.t
        self.ego_avg_speed += (1.0 - w) * (speed - self.ego_avg_speed)
        self.t = state.t

    def get_state(self) -> float:
        """Return the current average speed."""
        return self.ego_avg_speed


class EgoMaxSpeed(Metric):
    """Measure the maximum speed of the ego."""

    name = "ego_max_speed"

    def _reset(self, state: State) -> None:
        """Reset the maximum speed."""
        self.ego = state.scenario.ego
        self.ego_max_speed = np.linalg.norm(state.velocities[self.ego][:3])

    def _step(self, state: State) -> None:
        """Update the maximum speed."""
        speed = np.linalg.norm(state.velocities[self.ego][:3])
        self.ego_max_speed = np.maximum(speed, self.ego_max_speed)

    def get_state(self) -> float:
        """Return the current max speed."""
        return self.ego_max_speed


class EgoDistanceTravelled(Metric):
    """Measure the distance travelled by the ego."""

    name = "ego_distance_travelled"

    def _reset(self, state: State) -> None:
        """Find the ego."""
        self.ego = state.scenario.ego

    def _step(self, state: State) -> None:
        """Pass as entity will update its distance."""
        self.dist = state.distances[self.ego]

    def get_state(self) -> float:
        """Return the current distance travelled."""
        return self.dist



import inspect
import os
import warnings
from argparse import ArgumentParser
from typing import Any, Dict, List, Optional, Type, Union

import yaml

from scenario_gym.agent import Agent, ReplayTrajectoryAgent
from scenario_gym.controller import ReplayTrajectoryController
from scenario_gym.entity import Entity
from scenario_gym.metrics import Metric
from scenario_gym.scenario import Scenario
from scenario_gym.scenario_gym import ScenarioGym
from scenario_gym.sensor import EgoLocalizationSensor


def load_keywords(obj: Type, exclude: Optional[List[str]] = None) -> Dict[str, Any]:
    """Find keyword arguments of the object."""
    sig = inspect.signature(obj.__init__)
    return {
        k: v.default
        for k, v in sig.parameters.items()
        if v.default != inspect._empty and (exclude is None or k not in exclude)
    }


class ScenarioManager:
    """Provides functionality to manage running large numbers of scenarios."""

    GYM_PARAMETERS = load_keywords(ScenarioGym, exclude=["metrics"])
    PARAMETERS = {}

    @classmethod
    def generate_parser(cls) -> ArgumentParser:
        """Generate an argument parser for the manager."""
        parser = ArgumentParser(description=f"CLI for {cls.__name__}.")
        params = {
            **cls.GYM_PARAMETERS,
            **cls.VIEWER_PARAMETERS,
            **cls.PARAMETERS,
        }
        for k, v in params.items():
            if isinstance(v, bool):
                parser.add_argument(
                    f"--{k.replace(' ', '_')}",
                    action="store_false" if v else "store_true",
                )
            elif isinstance(v, (str, int, float)):
                parser.add_argument(
                    f"--{k.replace(' ', '_')}",
                    default=v,
                    type=type(v),
                )
            elif isinstance(v, (list, tuple)):
                parser.add_argument(
                    f"--{k.replace(' ', '_')}",
                    default=v,
                    nargs="+",
                )

            elif v is None:
                parser.add_argument(
                    f"--{k.replace(' ', '_')}",
                    default=v,
                    type=float,
                )
            else:
                warnings.warn(f"Type {type(v)} not supported.")
        return parser

    @classmethod
    def from_cli(cls, **kwargs):
        """Construct the manager from command line arguments."""
        parser = cls.generate_parser()
        args = parser.parse_args()
        return cls(
            **{
                k: v
                for k, v in args.__dict__.items()
                if k in cls.PARAMETERS and v is not None
            },
            **kwargs,
        )

    def __init__(
        self,
        config_path: Optional[str] = None,
        metrics: Optional[List[Metric]] = None,
        viewer_params: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """
        Construct the manager from input parameters.

        Parameters
        ----------
        config_path : Optional[str]
            A path to a yaml file of the parameters to be used.

        metrics : Optional[List[Metric]]
            List of metrics to measure.

        viewer_params : Optional[Dict[str, Any]]
            Dictionary of parameters to use for the Viewer. Since the viewer can be
            a custom class its parameters should be passed as a dictionary so the
            manager will know which parameters should be passed for it.

        **kwargs:
            Parameters given as keywords. Will override any parameters
            from a config file.

        """
        self.load_params(path=config_path, **kwargs)
        self.metrics = metrics.copy() if metrics is not None else []
        self.viewer_params = viewer_params.copy() if viewer_params else {}

    def load_params(self, config_path: Optional[str] = None, **kwargs) -> None:
        """Load all parameters required and set them as attributes."""
        params = yaml.safe_load(open(config_path, "r")) if config_path else {}
        self.PARAMETERS = self.PARAMETERS.copy()
        self.combined_config = {
            **self.GYM_PARAMETERS,
            **self.PARAMETERS,
            **params,
            **kwargs,
        }
        for k, v in self.combined_config.items():
            if not hasattr(self, k):
                setattr(self, k, v.copy() if isinstance(v, (list, dict)) else v)

    @property
    def parameter_names(self) -> List[str]:
        """Return the names of all parameters."""
        return self.gym_parameters_names + self.viewer_parameters_names

    @property
    def parameters(self) -> List[str]:
        """Return all the parameters for the gym and viewer."""
        return {**self.gym_parameters, **self.viewer_parameters}

    @property
    def gym_parameter_names(self) -> List[str]:
        """Return the names of all gym parameters."""
        return list(self.GYM_PARAMETERS)

    @property
    def gym_parameters(self) -> Dict[str, Any]:
        """Return the parameters needed for the ScenarioGym constructor."""
        return {k: getattr(self, k) for k in self.GYM_PARAMETERS}

    @property
    def viewer_parameter_names(self) -> List[str]:
        """Return the names of all viewer parameters."""
        return list(self.viewer_params)

    @property
    def viewer_parameters(self) -> Dict[str, Any]:
        """Return the parameters needed for the rendering module."""
        return self.viewer_params

    def make_gym(self, **kwargs) -> ScenarioGym:
        """Create a gym instance with the given config."""
        return ScenarioGym(
            metrics=self.metrics,
            **self.gym_parameters,
            **self.viewer_parameters,
            **kwargs,
        )

    def create_agent(self, scenario: Scenario, entity: Entity) -> Agent:
        """
        Construct the agents when loading a scenario.

        Parameters
        ----------
        scenario : Scenario
            The scenario object.

        entity : Entity
            The specific entity within the scenario.

        """
        if entity.ref == "ego":
            controller = ReplayTrajectoryController(entity)
            sensor = EgoLocalizationSensor(entity)
            return ReplayTrajectoryAgent(entity, controller, sensor)

    def add_metric(self, m: Metric) -> None:
        """Add a metric to the manager."""
        self.metrics.append(m)

    def on_rollout_start(self, gym: ScenarioGym) -> None:
        """Run before the rollout when running scenarios."""
        pass

    def on_rollout_end(self, gym: ScenarioGym) -> None:
        """Run after the rollout when running scenarios."""
        pass

    def run_scenario(
        self,
        scenario: Union[str, Scenario],
        render: bool = False,
        record: bool = False,
        **kwargs,
    ) -> List[Any]:
        """
        Run a single scenario in the gym.

        Parameters
        ----------
        scenario : Union[str, Scenario],
            The filepath of the OpenScenario file for the scenario or
            the scenario object.

        render : bool
            Whether to render the scenario.

        record : bool
            Whether to record the scenario to OpenScenario.

        """
        gym = self.make_gym()
        if record:
            gym.record()
        if isinstance(scenario, str):
            gym.load_scenario(scenario, create_agent=self.create_agent)
        elif isinstance(scenario, Scenario):
            gym.set_scenario(scenario, create_agent=self.create_agent)
        else:
            raise ValueError(f"{scenario}: should be a scenario or a file.")
        self.on_rollout_start(gym)
        gym.rollout(render=render, **kwargs)
        self.on_rollout_end(gym)
        if record:
            gym.recorder.get_state()
        return [m.get_state() for m in self.metrics]

    def run_scenarios(
        self,
        scenarios: List[str],
        render: bool = False,
        record: bool = False,
        **kwargs,
    ) -> List[List[Any]]:
        """
        Run a batch of scenarios.

        Parameters
        ----------
        scenarios : List[Union[str, Scenario]]
            The filepaths of the OpenScenario files for the scenarios or
            the raw scenario objects.

        render : bool
            Whether to render each scenario.

        record : bool
            Whether to record each scenario to OpenScenario.

        Returns
        -------
        List[List[Any]]
            The values for each metric after each scenario.

        """
        results = []
        gym = self.make_gym()
        if record:
            gym.record()
        for scenario in scenarios:
            if isinstance(scenario, str):
                gym.load_scenario(scenario, create_agent=self.create_agent)
            elif isinstance(scenario, Scenario):
                gym._set_scenario(scenario, create_agent=self.create_agent)
            else:
                raise ValueError(f"{scenario}: should be a scenario or a file.")
            gym.rollout(render=render, **kwargs)
            if record:
                gym.recorder.get_state()
            results.append([m.get_state() for m in self.metrics])
        return results

    def save_config(self, path: str = "./params.yml") -> None:
        """
        Write the config parameters to a yaml file.

        Parameters
        ----------
        path : str
            The filepath for the output.

        """
        path = os.path.splitext(path)[0] + ".yml"
        with open(path, "w") as f:
            yaml.dump(self.combined_config, f)



from abc import abstractmethod
from types import MethodType
from typing import Any, Callable, Optional, Tuple

from scenario_gym.agent import Agent
from scenario_gym.scenario_gym import ScenarioGym

try:
    from dm_env import Environment, TimeStep, restart, termination, transition
except ImportError:
    raise ImportError(
        "dm_env is required for this module. Install it with `pip install dm_env`."
    )


class ScenarioGym(ScenarioGym, Environment):
    """
    Scenario Gym subclass compatible with dm_env.

    This is still an abstract class which requires implementation of the method
    `observation_spec` and `action_spec`. This is because these are required as
    methods for `dm_env.Environment` but are not known by `ScenarioGym` until the
    ego agent is defined. Therefore subclasses should implement the specific specs
    required for the ego agent for the chosen experiment.
    """

    def __init__(
        self,
        *args,
        update_scenario: Optional[Callable[[ScenarioGym], None]] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.ego_agent: Optional[Agent] = None

        if update_scenario is not None:
            self.update_scenario = MethodType(update_scenario, self)

    def update_scenario(self) -> None:
        """Update the loaded scenario when reset is called."""
        pass

    def reset(self) -> TimeStep:
        """Reset the environment."""
        self.update_scenario()
        obs = self._reset()
        return restart(obs)

    def _reset(self) -> Any:
        """Reset the environment."""
        if self.state.scenario is None:
            raise ValueError("No scenario has been set.")
        self.reset_scenario()

        try:
            self.ego_agent = self.state.agents[self.state.scenario.ego]
        except KeyError:
            raise KeyError("No agent named ego.")

        self.state.next_t = self.state.t + self.timestep
        return self.ego_agent.sensor.step(self.state)

    def step(self, action) -> TimeStep:
        """Process an action and get the next timetsep."""
        if (
            self.state.scenario is None
            or self.state.is_done
            or self.ego_agent is None
        ):
            return self.reset()

        obs, reward = self._step(action)

        if self.state.is_done:
            return termination(reward, obs)

        return transition(reward, obs)

    def _step(self, action) -> Tuple[Any, float]:
        """Process the given action."""
        new_poses = {}
        for agent in self.state.agents.values():
            if agent is self.ego_agent:
                agent.last_action = action
                new_poses[agent.entity] = agent.controller.step(self.state, action)
            else:
                new_poses[agent.entity] = agent.step(self.state)
        new_poses.update(self.state.non_agents.step(self.state))

        # update the poses and current time
        self.state.step(new_poses)

        # get reward of next state
        reward = self.ego_agent.reward(self.state)

        # rendering and metrics
        for m in self.metrics:
            m.step(self.state)
        if self.viewer is not None:
            self.state.last_keystroke = self.render()

        # process ego part of next state
        self.state.next_t = self.state.t + self.timestep
        ego_obs = self.ego_agent.sensor.step(self.state)

        if self.state.is_done:
            for agent in self.state.scenario.agents.values():
                agent.finish(self.state)

        return ego_obs, reward

    @abstractmethod
    def observation_spec(self) -> Any:
        """Return the observation spec for the environment."""
        raise NotImplementedError()

    @abstractmethod
    def action_spec(self) -> Any:
        """Return the action spec for the environment."""
        raise NotImplementedError()

    def rollout(self, *args, **kwargs) -> None:
        """Raise an error if rollout is called with this env."""
        raise NotImplementedError("Rollout is not supported for this environment.")

    def reset_scenario(self) -> None:
        """Reset scenario and reference to old ego agent."""
        super().reset_scenario()
        self.ego_agent = None



from scenario_gym.xosc_interface.catalogs import read_catalog, write_catalogs
from scenario_gym.xosc_interface.read import import_scenario
from scenario_gym.xosc_interface.write import write_scenario



import numpy as np

from scenario_gym import ScenarioGym
from scenario_gym.viewer.opencv import OpenCVViewer


def test_viewer(all_scenarios):
    """Test that we can render with different options."""
    scenario_path = all_scenarios["a5e43fe4-646a-49ba-82ce-5f0063776566"]

    gym = ScenarioGym(
        timestep=0.1,
        viewer_class=OpenCVViewer,
        render_layers=list(OpenCVViewer._renderable_layers.keys()),
        **{
            f"{k}_color": tuple(map(int, np.random.randint(0, 255, 3)))
            for k in OpenCVViewer._renderable_layers
        },
        background=(0, 100, 0),
    )
    gym.load_scenario(scenario_path)
    for _ in range(1):
        gym.rollout(render=True)


def test_custom_viewer(all_scenarios):
    """Test rendering with a subclass of OpenCVViewer."""

    class CustomViewer(OpenCVViewer):
        """Viewer that changes colour for pedestrians and ego."""

        def get_entity_color(self, entity_idx: int, entity):
            if entity_idx == 0:
                return (0, 0, 240)  # red
            if entity.catalog_entry.catalog_category == "pedestrian":
                return (0, 240, 0)  # green
            return (200, 0, 0)  # blue

    scenario_path = all_scenarios["a5e43fe4-646a-49ba-82ce-5f0063776566"]

    gym = ScenarioGym(
        timestep=0.1,
        viewer_class=CustomViewer,
    )
    gym.load_scenario(scenario_path)
    gym.reset_scenario()
    gym.rollout(render=True)



import pytest as pt

from scenario_gym.callback import StateCallback
from scenario_gym.metrics import Metric
from scenario_gym.scenario_gym import ScenarioGym
from scenario_gym.trajectory import Trajectory
from scenario_gym.xosc_interface import import_scenario


@pt.fixture
def scenario(all_scenarios):
    """Get a scenario to test."""
    s = import_scenario(all_scenarios["3e39a079-5653-440c-bcbe-24dc9f6bf0e6"])
    data = s.entities[0].trajectory.data.copy()
    data[0, 0] = 0.0
    s.entities[0].trajectory = Trajectory(data)
    return s


class ExampleCallback(StateCallback):
    """A simple callback to add a feature to the state."""

    my_feature: bool

    def __call__(self, state):
        """Set the feature to whether the time is <= 1."""
        self.my_feature = state.t <= 1


class ExampleDependentCallback(StateCallback):
    """A simple callback that is dependented on the TestCallback."""

    required_callbacks = [ExampleCallback]

    dependent_feature: bool

    def __call__(self, state):
        """Set the feature to the opposite of the existing features."""
        self.dependent_feature = not self.callbacks[0].my_feature


def test_get_callback(scenario):
    """Test getting a callback instance from the state."""
    cb1, cb2 = ExampleCallback(), ExampleDependentCallback()
    gym = ScenarioGym(state_callbacks=[cb1, cb2])
    gym.set_scenario(scenario)
    assert gym.state.get_callback(ExampleCallback) is cb1
    assert gym.state.get_callback(ExampleDependentCallback) is cb2


def test_callback(scenario):
    """Test that we can add a callback to the state."""
    gym = ScenarioGym(state_callbacks=[ExampleCallback()])
    gym.set_scenario(scenario)
    assert gym.state_callbacks[0].my_feature, "Feature should be added at reset."
    gym.rollout()
    assert (
        gym.state.t > 1 and not gym.state_callbacks[0].my_feature
    ), "Feature should be False."


def test_metric_requires(scenario):
    """Test that metrics can be forced to require callbacks."""

    class ExampleMetric(Metric):
        required_callbacks = [ExampleCallback]

        def _reset(self, state):
            pass

        def _step(self, state):
            pass

        def get_state(self):
            pass

    gym = ScenarioGym(
        state_callbacks=[ExampleCallback()], metrics=[ExampleMetric()]
    )
    try:
        gym.set_scenario(scenario)
    except ValueError as e:
        raise "Should not raise any value error:" from e

    gym = ScenarioGym(state_callbacks=[], metrics=[ExampleMetric()])
    with pt.raises(ValueError):
        gym.set_scenario(scenario)


def test_callback_requires(scenario):
    """Test that the callbacks can be forced to require callbacks."""
    gym = ScenarioGym(
        state_callbacks=[ExampleCallback(), ExampleDependentCallback()]
    )
    try:
        gym.set_scenario(scenario)
    except ValueError as e:
        raise "Should not raise any value error:" from e

    gym = ScenarioGym(state_callbacks=[ExampleDependentCallback()])
    with pt.raises(ValueError):
        gym.set_scenario(scenario)

    gym = ScenarioGym(
        state_callbacks=[ExampleDependentCallback(), ExampleCallback()]
    )
    with pt.raises(ValueError):
        gym.set_scenario(scenario)



from dataclasses import dataclass
from typing import Any, Dict, Optional

from lxml.etree import Element
from scenariogeneration import xosc

from scenario_gym.catalog_entry import (
    ArgsKwargs,
    BoundingBox,
    Catalog,
    CatalogEntry,
)
from scenario_gym.entity.base import Entity
from scenario_gym.trajectory import Trajectory


@dataclass
class PedestrianCatalogEntry(CatalogEntry):
    """Catalog entry for a pedestrian."""

    mass: Optional[float]

    xosc_names = ["Pedestrian"]

    @classmethod
    def load_data_from_xml(
        cls,
        element: Element,
        catalog: Optional[Catalog] = None,
    ) -> ArgsKwargs:
        """Load the vehicle from an xml element."""
        base_args, _ = super().load_data_from_xml(element, catalog=catalog)
        mass = element.attrib.get("mass")
        if mass is not None:
            mass = float(mass)
        return base_args + (mass,), {}

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Load the pedestrian from a dictionary."""
        catalog = (
            Catalog.from_dict(data["catalog"])
            if data.get("catalog") is not None
            else None
        )
        return cls(
            catalog,
            data["catalog_entry"],
            data["catalog_category"],
            data["catalog_type"],
            BoundingBox.from_dict(data["bounding_box"]),
            data.get("properties", {}),
            data.get("files", []),
            data.get("mass"),
        )

    def to_dict(self) -> Dict[str, Any]:
        """Write the pedestrian to a dictionary."""
        data = super().to_dict()
        data["mass"] = self.mass
        return data

    def to_xosc(self) -> xosc.Pedestrian:
        """Write the pedestrian to xosc."""
        obj = xosc.Pedestrian(
            self.catalog_entry,
            self.mass,
            self.bounding_box.to_xosc(),
            getattr(
                xosc.PedestrianCategory,
                self.catalog_category,
                xosc.PedestrianCategory.pedestrian,
            ),
        )
        for k, v in self.properties.items():
            obj.add_property(k, v)
        for f in self.files:
            obj.add_property_file(f)
        return obj


class Pedestrian(Entity):
    """Entity class for pedestrians."""

    def __init__(
        self,
        catalog_entry: PedestrianCatalogEntry,
        trajectory: Optional[Trajectory] = None,
        ref: Optional[str] = None,
    ):
        super().__init__(catalog_entry, trajectory=trajectory, ref=ref)



from scenario_gym.action import Action
from scenario_gym.agent import Agent
from scenario_gym.catalog_entry import BoundingBox, CatalogEntry
from scenario_gym.controller import Controller
from scenario_gym.entity import Entity
from scenario_gym.manager import ScenarioManager
from scenario_gym.metrics import Metric
from scenario_gym.observation import Observation
from scenario_gym.road_network import RoadNetwork
from scenario_gym.scenario import Scenario
from scenario_gym.scenario_gym import ScenarioGym
from scenario_gym.sensor import Sensor
from scenario_gym.state import State
from scenario_gym.trajectory import Trajectory

__all__ = [
    # api objects
    "Action",
    "Agent",
    "BoundingBox",
    "CatalogEntry",
    "Controller",
    "ScenarioGym",
    "Entity",
    "Metric",
    "Observation",
    "RoadNetwork",
    "Scenario",
    "ScenarioManager",
    "Sensor",
    "State",
    "Trajectory",
    # modules
    "action",
    "agent",
    "catalog_entry",
    "controller",
    "scenario_gym",
    "entity",
    "manager",
    "metrics",
    "observation",
    "recorder",
    "road_network",
    "scenario",
    "sensor",
    "state",
    "trajectory",
    "viewer",
    "xosc_interface",
]

__version__ = "0.3.1"



import time

import pytest as pt

import scenario_gym
from scenario_gym.callback import StateCallback
from scenario_gym.scenario_gym import ScenarioGym
from scenario_gym.xosc_interface import import_scenario

speed_test = pt.mark.skipif("not config.getoption('speed_tests')")


@pt.fixture
def paths(all_scenarios):
    """Get filepaths for the scenarios to be used."""
    scenarios = [
        "41dac6fa-6f83-461e-a145-08692da5f3c7",
        "9c324146-be03-4d4e-8112-eaf36af15c17",
        "a5e43fe4-646a-49ba-82ce-5f0063776566",
        "a98d5c7d-76aa-49bf-b88c-97db5d5c7433",
        "d9726503-e04a-4e8b-b487-8805ef790c92",
        "e1bdb607-206b-4f40-9bc4-59ded182ecc8",
        "e56ae853-4266-4c30-865f-96737d87b601",
    ]
    paths = [all_scenarios[s] for s in scenarios]
    return paths


@speed_test
def test_gym_speed(paths):
    """
    Load and runs scenarios to test rollout speed..

    Use -s with pytest to see the result.
    """
    n = 10
    gym = ScenarioGym(timestep=1.0 / 30.0)
    start = time.time()
    for _ in range(n):
        for s in paths:
            gym.load_scenario(s)
            gym.rollout()
    t = (time.time() - start) / (n * len(paths))
    total_length = sum(import_scenario(p).length for p in paths)
    num_steps = [int(import_scenario(p).length * 30) for p in paths]
    print(
        "Completed in {:.4}s per scenario, {:.4}\u03BCs per step.".format(
            t, 1e6 * len(paths) * t / sum(num_steps)
        )
    )
    print("Running at {}x real time.".format(int(total_length / t)))


@speed_test
def test_collision_speed(paths):
    """
    Test the speed with collisions.

    Use -s with pytest to see the result.
    """

    class CollisionCallback(StateCallback):
        def __call__(self, state):
            state.collisions()

    n = 10
    gym = ScenarioGym(timestep=1.0 / 30.0, state_callbacks=[CollisionCallback()])
    start = time.time()
    for _ in range(n):
        for s in paths:
            gym.load_scenario(s)
            gym.rollout()
    t = (time.time() - start) / (n * len(paths))
    total_length = sum(import_scenario(p).length for p in paths)
    num_steps = [int(import_scenario(p).length * 30) for p in paths]
    print(
        "Completed in {:.4}s per scenario, {:.4}\u03BCs per step.".format(
            t, 1e6 * len(paths) * t / sum(num_steps)
        )
    )
    print("Running at {}x real time.".format(int(total_length / t)))


@speed_test
def test_render_speed(paths):
    """Load scenarios with rendering."""
    n = 2

    gym = ScenarioGym(timestep=1.0 / 30.0)
    start = time.time()
    for _ in range(n):
        for s in paths:
            gym.load_scenario(s)
            gym.rollout(render=True)
    t = (time.time() - start) / (n * len(paths))
    total_length = sum(import_scenario(p).length for p in paths)
    num_steps = [int(import_scenario(p).length * 30) for p in paths]
    print(
        "Completed in {:.4}s per scenario, {:.4}\u03BCs per step.".format(
            t, 1e6 * len(paths) * t / sum(num_steps)
        )
    )
    print("Running at {}x real time.".format(int(total_length / t)))


@speed_test
def test_sensor_speed(paths):
    """Measures speed when running with RasterizedMapSensor."""
    n = 10

    def create_agent(ref, entity):
        if ref == "ego":
            controller = scenario_gym.controller.ReplayTrajectoryController(entity)
            sensor = scenario_gym.sensor.RasterizedMapSensor(entity, freq=1)
            return scenario_gym.agent.ReplayTrajectoryAgent(
                entity, controller, sensor
            )

    gym = ScenarioGym(timestep=1.0 / 30.0)
    start = time.time()
    for _ in range(n):
        for s in paths:
            gym.load_scenario(s, create_agent=create_agent)
            gym.rollout()
    t = (time.time() - start) / (n * len(paths))
    num_steps = [int(import_scenario(p).length * 30) for p in paths]
    print(
        "Completed in {:.4}s per scenario, {:.4}s per step.".format(
            t, len(paths) * t / sum(num_steps)
        )
    )



from __future__ import annotations

from copy import copy
from inspect import getfullargspec
from typing import Any, Dict, Optional, Type

import numpy as np
from shapely.geometry import Polygon

from scenario_gym.catalog_entry import BoundingBox, CatalogEntry
from scenario_gym.trajectory import Trajectory
from scenario_gym.utils import ArrayLike, NDArray


class Entity:
    """
    An entity in the gym.

    An entity consists of a catalog entry and a pose. Note that poses
    are immutable arrays. Once a pose is set to the entity it cannot
    be changed, only overwritten. It could be modified e.g. by calling
    pose.copy() and then modifying the copied array and setting
    that back to the pose.
    """

    @classmethod
    def _catalog_entry_type(cls) -> Type[CatalogEntry]:
        """Get the type of catalog entry that is used for this entity."""
        args = getfullargspec(cls.__init__)
        ce = args.args[1]
        try:
            ce_type = args.annotations[ce]
        except KeyError as e:
            raise NotImplementedError(
                f"Subclass {cls.__name__} has no type annotation for catalog entry."
            ) from e
        if not issubclass(ce_type, CatalogEntry):
            raise TypeError("Catalog entry type must be a catalog entry subclass.")
        return ce_type

    def __init__(
        self,
        catalog_entry: CatalogEntry,
        trajectory: Optional[Trajectory] = None,
        ref: Optional[str] = None,
    ):
        """
        Construct an entity.

        Parameters
        ----------
        catalog_entry: CatalogEntry
            The catalog entry used for the entity.

        trajectory : trajectory
            The trajectory for the entity.

        ref : Optional[str]
            The unique reference for the entity from the OpenScenario file.

        """
        self.ref = ref
        self.catalog_entry = catalog_entry
        self._trajectory = trajectory

    @property
    def trajectory(self) -> Trajectory:
        """Get the trajectory for the entity."""
        return self._trajectory

    @trajectory.setter
    def trajectory(self, trajectory: Trajectory) -> None:
        """Set the trajectory for the entity."""
        self._trajectory = trajectory

    @property
    def bounding_box(self) -> BoundingBox:
        """Get the bounding box of the entity from its catalog entry."""
        return self.catalog_entry.bounding_box

    @property
    def type(self) -> Optional[str]:
        """Get the catalog type of the entity. E.g. Vehicle, Pedestrian."""
        return self.catalog_entry.catalog_type.replace("Catalogs", "")

    def __copy__(self) -> Entity:
        """Create a copy of an entity without copying the catalog_entry."""
        return self.__class__(
            self.catalog_entry,
            trajectory=None if self.trajectory is None else self.trajectory.copy(),
            ref=self.ref,
        )

    def copy(self) -> Entity:
        """Create a copy of an entity without copying the catalog_entry."""
        return copy(self)

    def get_bounding_box_points(self, pose: ArrayLike) -> NDArray:
        """
        Compute the bounding box coordinates in the global frame for the given pose.

        Returns in the order: RR, FR, FL, RL.

        Parameters
        ----------
        pose : Optional[ArrayLike]
            An array of the entities pose. May broadcast (..., [x, y, (z), h, ...]).

        """
        ref_xy, h = pose[..., :2], pose[..., 3 if pose.shape[-1] > 3 else 2]
        n = h.ndim
        R = np.array([[np.cos(h), np.sin(h)], [-np.sin(h), np.cos(h)]]).transpose(
            *(tuple(i + 2 for i in range(n)) + (0, 1))
        )
        points = np.array(
            [
                [
                    self.bounding_box.center_x - 0.5 * self.bounding_box.length,
                    self.bounding_box.center_y + 0.5 * self.bounding_box.width,
                ],
                [
                    self.bounding_box.center_x + 0.5 * self.bounding_box.length,
                    self.bounding_box.center_y + 0.5 * self.bounding_box.width,
                ],
                [
                    self.bounding_box.center_x + 0.5 * self.bounding_box.length,
                    self.bounding_box.center_y - 0.5 * self.bounding_box.width,
                ],
                [
                    self.bounding_box.center_x - 0.5 * self.bounding_box.length,
                    self.bounding_box.center_y - 0.5 * self.bounding_box.width,
                ],
            ]
        )
        points = ref_xy[..., None, :] + np.einsum("ij,...jk->...ik", points, R)
        return points

    def get_bounding_box_geom(self, pose: Optional[ArrayLike]) -> Polygon:
        """
        Return a Polygon representing the bounding box global frame.

        Returns cached values if the pose is None and they are there.

        Parameters
        ----------
        pose : Optional[ArrayLike]
            An array of the entities pose. Only one may be given.

        """
        return Polygon(self.get_bounding_box_points(pose))

    def is_static(self) -> bool:
        """Return True if the entity is static."""
        return self.trajectory.data.shape[0] == 1

    def to_dict(self) -> Dict[str, Any]:
        """Write the entity to a jsonable dictionary."""
        return {
            "ref": self.ref,
            "trajectory": self.trajectory.to_json(),
            "catalog_entry": self.catalog_entry.to_dict(),
            "entity_class": self.__class__.__name__,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Create the entity from the json dictionary."""
        return cls(
            cls._catalog_entry_type().from_dict(data["catalog_entry"]),
            trajectory=Trajectory(np.array(data["trajectory"])),
            ref=data.get("ref"),
        )


class StaticEntity(Entity):
    """Used for entities with only one control point."""

    @Entity.trajectory.setter
    def trajectory(self, trajectory: Trajectory) -> None:
        """Check that the trajectory is static."""
        if trajectory.data.shape[0] != 1:
            raise ValueError(
                "Recieved multiple control points for static entity: {self.ref}"
            )
        self._trajectory = trajectory



from scenario_gym.metrics.rss import RSS, RSSDistances
from scenario_gym.scenario_gym import ScenarioGym


def test_add_rss(all_scenarios):
    """Test adding the RSS metric to the scenario."""
    s = all_scenarios["3fee6507-fd24-432f-b781-ca5676c834ef"]
    gym = ScenarioGym(
        state_callbacks=[RSSDistances()],
        metrics=[RSS()],
    )
    gym.load_scenario(s)
    gym.rollout()

    data = gym.get_metrics()
    assert (
        "RSS_safe_longitudinal" in data and "RSS_safe_lateral" in data
    ), "Name did not get applied."
    assert len(data) == 2, "Incorrect number of metrics returned."
    assert (
        type(data["RSS_safe_longitudinal"]) is bool
    ), "Non-boolean RSS metric output for safe longitudinal distance."
    assert (
        type(data["RSS_safe_lateral"]) is bool
    ), "Non-boolean RSS metric output for safe lateral distance."
